"""
Configuration file for the Sphinx documentation builder.
"""
# -- stdlib imports ------------------------------------------------------------

import os
import sys
import datetime
import warnings
import tokenize
from pathlib import Path

from packaging.version import Version

# -- Read the Docs Specific Configuration --------------------------------------

# This needs to be done before sunpy is imported
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    os.environ['SUNPY_CONFIGDIR'] = '/home/docs/'
    os.environ['HOME'] = '/home/docs/'
    os.environ['LANG'] = 'C'
    os.environ['LC_ALL'] = 'C'
    os.environ['PARFIVE_HIDE_PROGRESS'] = 'True'

# -- Check for dependencies ----------------------------------------------------

from sunpy.util import missing_dependencies_by_extra

missing_requirements = missing_dependencies_by_extra("sunpy")["docs"]
if missing_requirements:
    print(
        f"The {' '.join(missing_requirements.keys())} package(s) could not be found and "
        "is needed to build the documentation, please install the 'docs' requirements."
    )
    sys.exit(1)

from matplotlib import MatplotlibDeprecationWarning
from ruamel.yaml import YAML
from sphinx_gallery.sorting import ExplicitOrder
from sunpy_sphinx_theme import PNG_ICON

from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.io.fits.verify import VerifyWarning
import sunpy
from sunpy.util.exceptions import SunpyDeprecationWarning, SunpyPendingDeprecationWarning

# -- Project information -------------------------------------------------------

# The full version, including alpha/beta/rc tags
from sunpy import __version__

_version = Version(__version__)
version = release = str(_version)
# Avoid "post" appearing in version string in rendered docs
if _version.is_postrelease:
    version = release = _version.base_version
# Avoid long githashes in rendered Sphinx docs
elif _version.is_devrelease:
    version = release = f"{_version.base_version}.dev{_version.dev}"
is_development = _version.is_devrelease
is_release = not(_version.is_prerelease or _version.is_devrelease)

project = "sunpy"
author = "The SunPy Community"
copyright = f'{datetime.datetime.now().year}, {author}'

# Register remote data option with doctest
import doctest

REMOTE_DATA = doctest.register_optionflag('REMOTE_DATA')

# We want to make sure all the following warnings fail the build on CI but not
# when actually building docs on RTD.
if not on_rtd:
    warnings.filterwarnings("error", category=SunpyDeprecationWarning)
    warnings.filterwarnings("error", category=SunpyPendingDeprecationWarning)
    warnings.filterwarnings("error", category=MatplotlibDeprecationWarning)
    warnings.filterwarnings("error", category=AstropyDeprecationWarning)
# Raised all by the sample data now and astropy 7,
# so we want to prevent this failing any of the builds
warnings.filterwarnings("ignore", category=VerifyWarning)

# -- SunPy Sample Data and Config ----------------------------------------------

# We set the logger to debug so that we can see any sample data download errors
# in the CI, especially RTD.
ori_level = sunpy.log.level
sunpy.log.setLevel("DEBUG")

import sunpy.data.sample

sunpy.data.sample.download_all()
sunpy.log.setLevel(ori_level)

# For the linkcheck
linkcheck_ignore = [
    r"https://doi\.org/\d+",
    r"https://\w\.element\.io/",
    # Checking all the PR URLs in the changelog takes a very long time
    r"https://github\.com/sunpy/sunpy/pull/\d+",
    # Avoid self-referencing links
    r"https://docs\.sunpy\.org",
    # Large PDF, so it is too slow to check
    r"https://inis\.iaea\.org/collection/NCLCollectionStore/_Public/20/062/20062491\.pdf",
    # These fails on SSL but are valid in a browser
    r"https://xrt\.cfa\.harvard\.edu/",
    r"https://opencv\.org",
    r"https://punch\.space\.swri\.edu",
    # This is super slow to check
    r"https://mathesaurus\.sourceforge\.net/idl-numpy\.html",
    # You have to be logged into GitHub in order to project wide issue searches
    r"https://github.com/issues?.*"
]
linkcheck_anchors = False
linkcheck_timeout = 120
linkcheck_workers = 10

# -- General configuration ---------------------------------------------------

# Wrap large function/method signatures
maximum_signature_line_length = 80
# sphinxext-opengraph
ogp_image = "https://raw.githubusercontent.com/sunpy/sunpy-logo/master/generated/sunpy_logo_word.png"
ogp_use_first_image = True
ogp_description_length = 160
ogp_custom_meta_tags = ('<meta property="og:ignore_canonical" content="true" />',)

# Suppress warnings about overriding directives as we overload some of the
# doctest extensions.
suppress_warnings = ['app.add_directive', ]

# Wrap large function/method signatures
maximum_signature_line_length = 80

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    'matplotlib.sphinxext.plot_directive',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'sphinx_changelog',
    'sphinx_gallery.gen_gallery',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sunpy.util.sphinx.doctest',
    'sunpy.util.sphinx.generate',
    "sphinxext.opengraph",
    'sphinx_design',
    'sphinx_copybutton',
    'sphinxcontrib.bibtex',
]

# Set automodapi to generate files inside the generated directory
automodapi_toctreedirnm = "generated/api"

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
html_extra_path = ['robots.txt']

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

if is_release:
    exclude_patterns.append('dev_guide/contents/*')

# The suffix(es) of source filenames.
# You can specify multiple suffix as a dict:
source_suffix = {".rst": "restructuredtext"}

# The master toctree document.
master_doc = 'index'

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = 'obj'

# Disable having a separate return type row
napoleon_use_rtype = False

# Disable google style docstrings
napoleon_google_docstring = False

# Disable the use of param, which prevents a distinct "Other Parameters" section
napoleon_use_param = False

# Enable nitpicky mode, which forces links to be non-broken
nitpicky = True
# This is not used. See docs/nitpick-exceptions file for the actual listing.
nitpick_ignore = []
for line in open('nitpick-exceptions'):
    if line.strip() == "" or line.startswith("#"):
        continue
    dtype, target = line.split(None, 1)
    target = target.strip()
    nitpick_ignore.append((dtype, target))


# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/python3.inv"),
    ),
    "numpy": (
        "https://numpy.org/doc/stable/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/numpy.inv"),
    ),
    "scipy": (
        "https://docs.scipy.org/doc/scipy/reference/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/scipy.inv"),
    ),
    "aiapy": ("https://aiapy.readthedocs.io/en/stable/", None),
    "asdf": ("https://asdf.readthedocs.io/en/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "dask": ("https://docs.dask.org/en/stable/", None),
    "drms": ("https://docs.sunpy.org/projects/drms/en/stable/", None),
    "hvpy": ("https://hvpy.readthedocs.io/en/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "mpl_animators": ("https://docs.sunpy.org/projects/mpl-animators/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "parfive": ("https://parfive.readthedocs.io/en/stable/", None),
    "reproject": ("https://reproject.readthedocs.io/en/stable/", None),
    "skimage": ("https://scikit-image.org/docs/stable/", None),
    "spiceypy": ("https://spiceypy.readthedocs.io/en/stable/", None),
    "sunkit_image": ("https://docs.sunpy.org/projects/sunkit-image/en/stable/", None),
    "sunkit_instruments": ("https://docs.sunpy.org/projects/sunkit-instruments/en/stable/", None),
    "zeep": ("https://docs.python-zeep.org/en/stable/", None),
    "contourpy": ("https://contourpy.readthedocs.io/en/stable/", None),
    "sphinxcontrib_bibtex": ("https://sphinxcontrib-bibtex.readthedocs.io/en/stable/", None),
}

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation for
# a list of builtin themes.
html_theme = "sunpy"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# By default, when rendering docstrings for classes, sphinx.ext.autodoc will
# make docs with the class-level docstring and the class-method docstrings,
# but not the __init__ docstring, which often contains the parameters to
# class constructors across the scientific Python ecosystem. The option below
# will append the __init__ docstring to the class-level docstring when rendering
# the docs. For more options, see:
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autoclass_content
autoclass_content = "both"

bibtex_bibfiles = ['references.bib']

# -- Linking to source code ----------------------------------------------------

link_github = True
# You can add build old with link_github = False

if link_github:
    import inspect

    extensions.append('sphinx.ext.linkcode')

    def linkcode_resolve(domain, info):
        """
        Determine the URL corresponding to Python object
        """
        if domain != 'py':
            return None

        modname = info['module']
        fullname = info['fullname']

        submod = sys.modules.get(modname)
        if submod is None:
            return None

        obj = submod
        for part in fullname.split('.'):
            try:
                obj = getattr(obj, part)
            except AttributeError:
                return None

        if inspect.isfunction(obj):
            obj = inspect.unwrap(obj)
        try:
            fn = inspect.getsourcefile(obj)
        except TypeError:
            fn = None
        if not fn or fn.endswith('__init__.py'):
            try:
                fn = inspect.getsourcefile(sys.modules[obj.__module__])
            except (TypeError, AttributeError, KeyError):
                fn = None
        if not fn:
            return None

        try:
            source, lineno = inspect.getsourcelines(obj)
        except (OSError, TypeError, tokenize.TokenError):
            if hasattr(obj, '__qualname__'):
                print(f"linkcode_resolve: could not get source for {obj.__module__}.{obj.__qualname__}")
            lineno = None

        linespec = (f"#L{lineno:d}-L{lineno + len(source) - 1:d}"
                    if lineno else "")

        startdir = Path(sunpy.__file__).parent.parent
        try:
            fn = os.path.relpath(fn, start=startdir).replace(os.path.sep, '/')
        except ValueError:
            return None

        if not fn.startswith('sunpy/'):
            return None

        tag = 'main' if is_development else f'v{_version.public}'
        return (f"https://github.com/sunpy/sunpy/blob/{tag}/{fn}{linespec}")
else:
    extensions.append('sphinx.ext.viewcode')

# -- Sphinx Gallery ------------------------------------------------------------

# JSOC email os env
# see https://github.com/sunpy/sunpy/wiki/Home:-JSOC
os.environ["JSOC_EMAIL"] = "jsoc@sunpy.org"
sphinx_gallery_conf = {
    'backreferences_dir': os.path.join('generated', 'modules'),
    'filename_pattern': '^((?!skip_).)*$',
    'examples_dirs': os.path.join('..', 'examples'),
    'subsection_order': ExplicitOrder([
        '../examples/acquiring_data',
        '../examples/map',
        '../examples/map_transformations',
        '../examples/time_series',
        '../examples/units_and_coordinates',
        '../examples/plotting',
        '../examples/differential_rotation',
        '../examples/saving_and_loading_data',
        '../examples/computer_vision_techniques',
        '../examples/showcase',
    ]),
    'within_subsection_order': "ExampleTitleSortKey",
    'gallery_dirs': os.path.join('generated', 'gallery'),
    'matplotlib_animations': True,
    # Comes from the theme.
    "default_thumb_file": PNG_ICON,
    'abort_on_example_error': False,
    'plot_gallery': 'True',
    'remove_config_comments': True,
    'doc_module': ('sunpy'),
    'only_warn_on_example_error': True,
    'parallel': True,
}

# -- Linking to OpenCV docs by using rst_epilog --------------------------------

try:
    import requests
    from bs4 import BeautifulSoup

    base_url = "https://docs.opencv.org"

    # The stable-version docs are the first item in the second list on the main page
    all_docs = BeautifulSoup(requests.get(base_url).text, 'html.parser')
    version = all_docs.find_all('ul')[1].li.a.attrs['href'][2:]  # strip leading "./"

    # Find the relative URL to the page for the `cv` namespace
    stable_docs = BeautifulSoup(requests.get(f"{base_url}/{version}/namespaces.html").text,
                                'html.parser')
    cv_namespace = stable_docs.find("a", string="cv").attrs['href']

    # Find the relative URL for warpAffine/filter2D in the `cv` namespace
    all_cv = BeautifulSoup(requests.get(f"{base_url}/{version}/{cv_namespace}").text, 'html.parser')
    warpAffine = all_cv.find("a", string="warpAffine").attrs['href'][6:]  # strip leading "../../"
    filter2D = all_cv.find("a", string="filter2D").attrs['href'][6:]  # strip leading "../../"

    # Construct the full URL for warpAffine/filter2D
    warpAffine_full = f"{base_url}/{version}/{warpAffine}"
    filter2D_full = f"{base_url}/{version}/{filter2D}"
except Exception:
    # In the event of any failure (e.g., no network connectivity)
    warpAffine_full = ""
    filter2D_full = ""

rst_epilog = f"""
.. |cv2_warpAffine| replace:: **cv2.warpAffine()**
.. _cv2_warpAffine: {warpAffine_full}
.. |cv2_filter2D| replace:: **cv2.filter2D()**
.. _cv2_filter2D: {filter2D_full}
"""

# -- Options for sphinx-copybutton ---------------------------------------------

# Python Repl + continuation, Bash, ipython and qtconsole + continuation, jupyter-console + continuation
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Stability Page ------------------------------------------------------------

with open('./reference/sunpy_stability.yaml') as estability:
    yaml = YAML(typ='rt')
    sunpy_modules = yaml.load(estability.read())

html_context = {
    'sunpy_modules': sunpy_modules,
    'is_development': not is_release,
}


def jinja_to_rst(app, docname, source):
    """
    Render our pages as a jinja template for fancy templating goodness.

    Depending on the building format, we render the page as a jinja template.

    For the linkchecker, we bypass templating as it doesn't support jinja, but we
    remove the jinja blocks so the page is still valid for checking.
    """
    jinja_pages = ["reference/stability", "dev_guide/index"]
    if app.builder.format == 'html':
        if docname in jinja_pages:
            print(f"Jinja rendering {docname}")
            rendered = app.builder.templates.render_string(
                source[0], app.config.html_context
            )
            source[0] = rendered
    else:
        if docname == "dev_guide/index":
            # This page only has a single jinja block that renders if the docs are
            # being built in development mode. We can simply  remove this block.
            for to_replace in ["{% if is_development %}", "{%else%}", "{% endif %}"]:
                source[0] = source[0].replace(to_replace, "")
        if docname == "reference/stability":
            # This page is a bit more complex, so we will remove the entire jinja block
            # leaving on the starting text of the page. Luckily there are no URLs in the
            # jinja block so this is safe.
            source[0] = source[0].split("{")[0]


# -- Sphinx setup --------------------------------------------------------------

def setup(app):
    # Handles the templating for the jinja pages in our docs
    app.connect("source-read", jinja_to_rst)



import re

import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import BaseCoordinateFrame, SkyCoord
from astropy.tests.helper import assert_quantity_allclose

import sunpy.map
from sunpy.coordinates import HeliographicStonyhurst
from sunpy.coordinates.frames import HeliographicCarrington
from sunpy.coordinates.utils import GreatArc
from sunpy.map.maputils import (
    _verify_coordinate_helioprojective,
    all_coordinates_from_map,
    all_corner_coords_from_map,
    all_pixel_indices_from_map,
    contains_coordinate,
    contains_full_disk,
    contains_limb,
    contains_solar_center,
    coordinate_is_on_solar_disk,
    is_all_off_disk,
    is_all_on_disk,
    map_edges,
    on_disk_bounding_coordinates,
    pixelate_coord_path,
    sample_at_coords,
    solar_angular_radius,
)


@pytest.fixture
def all_off_disk_map(aia171_test_map):
    return aia171_test_map.submap((1, 1)*u.pix, top_right=(11, 12)*u.pix)


@pytest.fixture
def all_on_disk_map(aia171_test_map):
    return aia171_test_map.submap((30, 60)*u.pix, top_right=(50, 85)*u.pix)


@pytest.fixture
def straddles_limb_map(aia171_test_map):
    return aia171_test_map.submap((64, 80)*u.pix, top_right=(120, 127)*u.pix)


@pytest.fixture
def sub_smap(aia171_test_map):
    return aia171_test_map.submap((0, 0)*u.pix, top_right=(50, 60)*u.pix)


@pytest.fixture
def non_helioprojective_map():
    data = np.arange(0, 100).reshape(10, 10)
    coord = SkyCoord(0*u.arcsec, 0*u.arcsec, obstime='2013-10-28 08:24',
                     observer='earth', frame=HeliographicCarrington)
    header = sunpy.map.header_helper.make_fitswcs_header(data, coord,
                                                         reference_pixel=[0, 0]*u.pixel,
                                                         scale=[2, 2]*u.arcsec/u.pixel,
                                                         telescope='Fake Telescope', instrument='UV detector',
                                                         wavelength=1000*u.angstrom)
    return sunpy.map.Map(data, header)


@pytest.fixture
def non_helioprojective_skycoord():
    return SkyCoord(0 * u.rad, 0 * u.rad, frame="icrs")


@pytest.fixture
def aia_test_arc(aia171_test_map):
    start = SkyCoord(735 * u.arcsec, -471 * u.arcsec, frame=aia171_test_map.coordinate_frame)
    end = SkyCoord(-100 * u.arcsec, 800 * u.arcsec, frame=aia171_test_map.coordinate_frame)
    return GreatArc(start, end)


def test_all_pixel_indices_from_map(sub_smap):
    pixel_indices = all_pixel_indices_from_map(sub_smap)
    shape = sub_smap.data.shape
    ny = shape[0]
    nx = shape[1]
    assert np.all(pixel_indices.shape == (2, ny, nx))
    assert np.all(pixel_indices.unit == u.pix)
    assert np.all(pixel_indices[:, 0, 0] == [0., 0.] * u.pix)
    assert np.all(pixel_indices[:, 0, nx-1] == [nx-1, 0.] * u.pix)
    assert np.all(pixel_indices[:, ny-1, 0] == [0., ny-1] * u.pix)
    assert np.all(pixel_indices[:, ny-1, nx-1] == [nx-1, ny-1] * u.pix)


def test_all_coordinates_from_map(sub_smap):
    coordinates = all_coordinates_from_map(sub_smap)
    shape = sub_smap.data.shape
    assert coordinates.shape == (shape[0], shape[1])
    assert isinstance(coordinates, SkyCoord)
    assert isinstance(coordinates.frame, BaseCoordinateFrame)
    assert coordinates.frame.name == sub_smap.coordinate_frame.name

    xpix, ypix = sub_smap.world_to_pixel(coordinates[0, 0])
    assert_quantity_allclose(xpix, 0*u.pix, atol=1e-7*u.pix)
    assert_quantity_allclose(ypix, 0*u.pix, atol=1e-7*u.pix)

    xpix, ypix = sub_smap.world_to_pixel(coordinates[-1, -1])
    assert_quantity_allclose(xpix, sub_smap.dimensions[0] - 1*u.pix)
    assert_quantity_allclose(ypix, sub_smap.dimensions[1] - 1*u.pix)


def test_all_corner_coordinates_from_map(sub_smap):
    coordinates = all_corner_coords_from_map(sub_smap)
    shape = sub_smap.data.shape
    assert coordinates.shape == (shape[0] + 1, shape[1] + 1)
    assert isinstance(coordinates, SkyCoord)
    assert isinstance(coordinates.frame, BaseCoordinateFrame)
    assert coordinates.frame.name == sub_smap.coordinate_frame.name

    xpix, ypix = sub_smap.world_to_pixel(coordinates[0, 0])
    assert_quantity_allclose(xpix, -0.5*u.pix)
    assert_quantity_allclose(ypix, -0.5*u.pix)

    xpix, ypix = sub_smap.world_to_pixel(coordinates[-1, -1])
    assert_quantity_allclose(xpix, sub_smap.dimensions[0] - 0.5*u.pix)
    assert_quantity_allclose(ypix, sub_smap.dimensions[1] - 0.5*u.pix)


def test_map_edges(all_off_disk_map):
    edges = map_edges(all_off_disk_map)
    assert isinstance(edges, tuple)
    assert len(edges[2]) == 12
    assert np.all(edges[2][0] == [0, 0] * u.pix)
    assert np.all(edges[2][11] == [0, 11] * u.pix)

    assert len(edges[3]) == 12
    assert np.all(edges[3][0] == [10, 0] * u.pix)
    assert np.all(edges[3][11] == [10, 11] * u.pix)

    assert len(edges[1]) == 11
    assert np.all(edges[1][0] == [0, 0] * u.pix)
    assert np.all(edges[1][10] == [10, 0] * u.pix)

    assert len(edges[0]) == 11
    assert np.all(edges[0][0] == [0, 11] * u.pix)
    assert np.all(edges[0][10] == [10, 11] * u.pix)


def test_solar_angular_radius(aia171_test_map):
    on_disk = aia171_test_map.center
    sar = solar_angular_radius(on_disk)
    assert isinstance(sar, u.Quantity)
    np.testing.assert_almost_equal(sar.to(u.arcsec).value, 971.80181131, decimal=1)


def test_contains_full_disk(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map):
    assert contains_full_disk(aia171_test_map)
    assert ~contains_full_disk(all_off_disk_map)
    assert ~contains_full_disk(all_on_disk_map)
    assert ~contains_full_disk(straddles_limb_map)


def test_is_all_off_disk(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map):
    assert ~is_all_off_disk(aia171_test_map)
    assert is_all_off_disk(all_off_disk_map)
    assert ~is_all_off_disk(all_on_disk_map)
    assert ~is_all_off_disk(straddles_limb_map)


def test_is_all_on_disk(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map):
    assert ~is_all_on_disk(aia171_test_map)
    assert ~is_all_on_disk(all_off_disk_map)
    assert is_all_on_disk(all_on_disk_map)
    assert ~is_all_on_disk(straddles_limb_map)


def test_contains_limb(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map):
    assert contains_limb(aia171_test_map)
    assert ~contains_limb(all_off_disk_map)
    assert ~contains_limb(all_on_disk_map)
    assert contains_limb(straddles_limb_map)


def test_coordinate_is_on_solar_disk(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map):
    off_disk = aia171_test_map.bottom_left_coord
    on_disk = aia171_test_map.center

    # Check for individual coordinates
    assert coordinate_is_on_solar_disk(on_disk)
    assert ~coordinate_is_on_solar_disk(off_disk)

    # Raise the error
    with pytest.raises(ValueError, match=re.escape("The input coordinate(s) is of type HeliographicStonyhurst, but must be in the Helioprojective frame.")):
        coordinate_is_on_solar_disk(on_disk.transform_to(HeliographicStonyhurst))

    # Check for sets of coordinates
    assert np.any(coordinate_is_on_solar_disk(all_coordinates_from_map(aia171_test_map)))
    assert np.any(~coordinate_is_on_solar_disk(all_coordinates_from_map(aia171_test_map)))
    assert np.all(~coordinate_is_on_solar_disk(all_coordinates_from_map(all_off_disk_map)))
    assert np.all(coordinate_is_on_solar_disk(all_coordinates_from_map(all_on_disk_map)))
    assert np.any(coordinate_is_on_solar_disk(all_coordinates_from_map(straddles_limb_map)))
    assert np.any(~coordinate_is_on_solar_disk(all_coordinates_from_map(straddles_limb_map)))


# Testing values are derived from running the code, not from external sources
def test_on_disk_bounding_coordinates(aia171_test_map):
    bl, tr = on_disk_bounding_coordinates(aia171_test_map)
    np.testing.assert_almost_equal(bl.Tx.to(u.arcsec).value, -954.17124289, decimal=1)
    np.testing.assert_almost_equal(bl.Ty.to(u.arcsec).value, -965.93063472, decimal=1)
    np.testing.assert_almost_equal(tr.Tx.to(u.arcsec).value, 964.27061417, decimal=1)
    np.testing.assert_almost_equal(tr.Ty.to(u.arcsec).value, 971.63586861, decimal=1)


def test_data_at_coordinates(aia171_test_map, aia_test_arc):
    data = sample_at_coords(aia171_test_map, aia_test_arc.coordinates())
    pixels = np.asarray(np.rint(
        aia171_test_map.world_to_pixel(aia_test_arc.coordinates())), dtype=int)
    x = pixels[0, :]
    y = pixels[1, :]
    intensity_along_arc = aia171_test_map.data[y, x] * aia171_test_map.unit
    assert_quantity_allclose(data[0], intensity_along_arc[0])
    assert_quantity_allclose(data[-1], intensity_along_arc[-1])


def test_sample_out_of_bounds(aia171_test_map):
    point = aia171_test_map.pixel_to_world([-1, 1]*u.pix, [-1, 1]*u.pix)
    with pytest.raises(ValueError, match='At least one coordinate is not within the bounds of the map.'):
        sample_at_coords(aia171_test_map, point)


def test_contains_solar_center(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map, sub_smap):
    assert contains_solar_center(aia171_test_map)
    assert not contains_solar_center(all_off_disk_map)
    assert not contains_solar_center(all_on_disk_map)
    assert not contains_solar_center(straddles_limb_map)
    assert not contains_solar_center(sub_smap)


def test_verify_coordinate_helioprojective(aia171_test_map, all_off_disk_map, all_on_disk_map, straddles_limb_map,
                                           sub_smap, non_helioprojective_map, non_helioprojective_skycoord):
    # These should be helioprojective.
    _verify_coordinate_helioprojective(aia171_test_map.coordinate_frame)
    _verify_coordinate_helioprojective(all_off_disk_map.coordinate_frame)
    _verify_coordinate_helioprojective(all_on_disk_map.coordinate_frame)
    _verify_coordinate_helioprojective(straddles_limb_map.coordinate_frame)
    _verify_coordinate_helioprojective(sub_smap.coordinate_frame)
    # These are not.
    with pytest.raises(ValueError, match=r"HeliographicCarrington, .* Helioprojective"):
        _verify_coordinate_helioprojective(non_helioprojective_map.coordinate_frame)
    with pytest.raises(ValueError, match=r"ICRS, .* Helioprojective"):
        _verify_coordinate_helioprojective(non_helioprojective_skycoord)


def test_functions_raise_non_frame_coords(non_helioprojective_skycoord):
    with pytest.raises(ValueError, match=r"ICRS, .* Helioprojective"):
        solar_angular_radius(non_helioprojective_skycoord)
    with pytest.raises(ValueError, match=r"ICRS, .* Helioprojective"):
        coordinate_is_on_solar_disk(non_helioprojective_skycoord)


def test_functions_raise_non_frame_map(non_helioprojective_map):
    with pytest.raises(ValueError, match=r"HeliographicCarrington, .* Helioprojective"):
        contains_full_disk(non_helioprojective_map)
    with pytest.raises(ValueError, match=r"HeliographicCarrington, .* Helioprojective"):
        contains_solar_center(non_helioprojective_map)
    with pytest.raises(ValueError, match=r"HeliographicCarrington, .* Helioprojective"):
        is_all_off_disk(non_helioprojective_map)
    with pytest.raises(ValueError, match=r"HeliographicCarrington, .* Helioprojective"):
        contains_limb(non_helioprojective_map)
    with pytest.raises(ValueError, match=r"HeliographicCarrington, .* Helioprojective"):
        on_disk_bounding_coordinates(non_helioprojective_map)


def test_contains_coord(aia171_test_map):
    smap = aia171_test_map
    for coord in [smap.bottom_left_coord,
                  smap.top_right_coord,
                  SkyCoord(0*u.deg, 0*u.deg, frame=smap.coordinate_frame)]:
        assert contains_coordinate(smap, coord)

    assert not contains_coordinate(smap, SkyCoord(2000*u.arcsec, 2000*u.arcsec,
                                                  frame=smap.coordinate_frame))

    multi_coord = SkyCoord([0, 2000]*u.arcsec, [0, 2000]*u.arcsec,
                           frame=smap.coordinate_frame)
    assert (contains_coordinate(smap, multi_coord) == [True, False]).all()


@pytest.mark.parametrize(('x', 'y', 'sampled_x', 'sampled_y'),
                         [([1, 5], [1, 1], [1, 2, 3, 4, 5], [1, 1, 1, 1, 1]),
                          ([1, 5], [1, 2], [1, 2, 3, 3, 4, 5], [1, 1, 1, 2, 2, 2]),
                          ([1, 5], [1, 3], [1, 2, 2, 3, 4, 4, 5], [1, 1, 2, 2, 2, 3, 3]),
                          ([1, 5], [1.0, 4.0], [1, 2, 2, 3, 3, 4, 4, 5], [1, 1, 2, 2, 3, 3, 4, 4]),
                          ([1, 5], [1.0, 3.6], [1, 2, 2, 3, 3, 4, 5, 5], [1, 1, 2, 2, 3, 3, 3, 4]),
                          ([1, 5], [1.4, 3.6], [1, 1, 2, 3, 3, 4, 5, 5], [1, 2, 2, 2, 3, 3, 3, 4])])
def test_pixelate_coord_path(aia171_test_map, x, y, sampled_x, sampled_y):
    # Also test the x<->y transpose
    for xx, yy, sxx, syy in [(x, y, sampled_x, sampled_y), (y, x, sampled_y, sampled_x)]:
        # Using the AIA test map for a "real" WCS, but the actual WCS is irrelevant for this test
        line = aia171_test_map.wcs.pixel_to_world(xx, yy)
        sampled_coords = pixelate_coord_path(aia171_test_map, line)
        sampled_pixels = aia171_test_map.wcs.world_to_pixel(sampled_coords)
        assert np.allclose(sampled_pixels[0], sxx)
        assert np.allclose(sampled_pixels[1], syy)


@pytest.mark.parametrize(('x', 'y', 'sampled_x', 'sampled_y'),
                         [([1, 5], [1, 1], [1, 2, 3, 4, 5], [1, 1, 1, 1, 1]),
                          ([1, 5], [1, 2], [1, 2, 3, 4, 5], [1, 1, 1, 2, 2]),
                          ([1, 5], [1, 3], [1, 2, 3, 4, 5], [1, 1, 2, 2, 3]),
                          ([1, 5], [1.0, 4.0], [1, 2, 3, 4, 5], [1, 2, 2, 3, 4]),
                          ([1, 5], [1.0, 3.6], [1, 2, 3, 4, 5], [1, 2, 2, 3, 4]),
                          ([1, 5], [1.4, 3.6], [1, 2, 3, 4, 5], [1, 2, 2, 3, 4])])
def test_pixelate_coord_path_bresenham(aia171_test_map, x, y, sampled_x, sampled_y):
    # Also test the x<->y transpose
    for xx, yy, sxx, syy in [(x, y, sampled_x, sampled_y), (y, x, sampled_y, sampled_x)]:
        # Using the AIA test map for a "real" WCS, but the actual WCS is irrelevant for this test
        line = aia171_test_map.wcs.pixel_to_world(xx, yy)
        sampled_coords = pixelate_coord_path(aia171_test_map, line, bresenham=True)
        sampled_pixels = aia171_test_map.wcs.world_to_pixel(sampled_coords)
        assert np.allclose(sampled_pixels[0], sxx)
        assert np.allclose(sampled_pixels[1], syy)



"""
This module provides a web scraper.
"""
import os
import re
from time import sleep
from ftplib import FTP
from datetime import datetime
from urllib.error import URLError, HTTPError
from urllib.parse import urlsplit
from urllib.request import urlopen

from bs4 import BeautifulSoup

from astropy.time import Time

from sunpy import log
from sunpy.extern.parse import parse
from sunpy.net.scraper_utils import date_floor, extract_timestep, get_timerange_from_exdict

__all__ = ['Scraper']


# Regular expressions to convert datetime format
TIME_CONVERSIONS = {
    '%Y': r'\d{4}', '%y': r'\d{2}',
    '%b': '[A-Z][a-z]{2}', '%B': r'\W', '%m': r'\d{2}',
    '%d': r'\d{2}', '%j': r'\d{3}',
    '%H': r'\d{2}', '%I': r'\d{2}',
    '%M': r'\d{2}',
    '%S': r'\d{2}', '%e': r'\d{3}', '%f': r'\d{6}'
}
# "parse" expressions to convert into datetime format
PARSE_TIME_CONVERSIONS = {
    "{year:4d}": "%Y", "{year:2d}": "%y",
    "{month:2d}": "%m",
    "{month_name:l}": "%B",
    "{month_name_abbr:l}": "%b",
    "{day:2d}": "%d", "{day_of_year:3d}": "%j",
    "{hour:2d}": "%H",
    "{minute:2d}": "%M",
    "{second:2d}": "%S",
    "{microsecond:6d}": "%f",
    "{millisecond:3d}": "%e",
    "{week_number:2d}": "%W",
}

class Scraper:
    """
    A scraper to scrap web data archives based on dates.

    Parameters
    ----------
    format : `str`
        A string containing the url with the date and other information to be
        extracted encoded as ``parse`` formats, and any other ``kwargs`` parameters
        as a string format, the former represented using double curly-brackets
        to differentiate from the latter.
        The accepted parse representations for datetime values are as given in ``PARSE_TIME_CONVERSIONS``.
        This can also be a uri to a local file patterns. Default is `None`.
    kwargs : `dict`
        A dictionary containing the values to be replaced in the pattern.
        Will be ignored if ``regex`` is `True`.

    Attributes
    ----------
    pattern : `str`
        The pattern with the parse format.
    datetime_pattern : `str`
        The parse pattern in the datetime format.
    now : `datetime.datetime`
        The pattern with the actual date.
        This is not checking if there is an existent file, but just how the ``pattern`` looks with the current time.

    Examples
    --------
    >>> from sunpy.net import Scraper
    >>>
    >>> pattern = ('https://proba2.sidc.be/{instrument}/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/'
    ...            '{instrument}_lv1_{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{month:2d}}{{second:2d}}.fits')
    >>> swap = Scraper(format=pattern, instrument='swap')
    >>>
    >>> print(swap.pattern)
    https://proba2.sidc.be/swap/data/bsd/{year:4d}/{month:2d}/{day:2d}/swap_lv1_{year:4d}{month:2d}{day:2d}_{hour:2d}{month:2d}{second:2d}.fits
    >>>
    >>> print(swap.datetime_pattern)
    https://proba2.sidc.be/swap/data/bsd/%Y/%m/%d/swap_lv1_%Y%m%d_%H%m%S.fits
    >>>
    >>> print(swap.now)  # doctest: +SKIP
    https://proba2.sidc.be/swap/data/bsd/2022/12/21/swap_lv1_20221221_112433.fits
    """
    def __init__(self, format, **kwargs):
        pattern = format.format(**kwargs)
        datetime_pattern = pattern
        for k, v in PARSE_TIME_CONVERSIONS.items():
            if k in datetime_pattern:
                datetime_pattern = datetime_pattern.replace(k,v)
        self.datetime_pattern = datetime_pattern
        # so that they don't conflict later on, either one can help in extracting year
        if "year:4d" in pattern and "year:2d" in pattern:
            pattern = pattern.replace("year:2d", ":2d")
        self.pattern = pattern
        self.domain = f"{urlsplit(self.pattern).scheme}://{urlsplit(self.pattern).netloc}/"
        milliseconds = re.search(r'\%e', self.datetime_pattern)
        if not milliseconds:
            self.now = datetime.now().strftime(self.datetime_pattern)
        else:
            now = datetime.now()
            milliseconds_ = int(now.microsecond / 1000.)
            self.now = now.strftime(f'{self.datetime_pattern[0:milliseconds.start()]}{milliseconds_:03d}{self.datetime_pattern[milliseconds.end():]}')

    def matches(self, filepath, date):
        """
        Checks if the given filepath is how the file path is expected
        to look on given date based on the pattern.

        Parameters
        ----------
        filepath : `str`
            File path to check.
        date : `datetime.datetime` or `astropy.time.Time`
            The date for which to check.

        Returns
        -------
        `bool`
            `True` if the given filepath matches with the calculated one for given date, else `False`.
        """
        return parse(date.strftime(self.datetime_pattern), filepath) is not None

    def range(self, timerange):
        """
        Gets the directories for a certain range of time.

        Parameters
        ----------
        timerange : `~sunpy.time.timerange.TimeRange`
            Time interval where to find the directories for a given pattern.

        Returns
        -------
        `list` of `str`
            All the possible directories valid for the time range given.
            Notice that these directories may not exist in the archive.
        """
        # find directory structure - without file names
        if '/' in self.datetime_pattern:
            directorypattern = '/'.join(self.datetime_pattern.split('/')[:-1]) + '/'
        timestep = extract_timestep(directorypattern)
        if timestep is None:
            return [directorypattern]
        else:
            directories = []
            cur = date_floor(timerange.start, timestep)
            end = date_floor(timerange.end, timestep) + timestep
            while cur < end:
                directories.append(cur.strftime(directorypattern))
                cur = cur + timestep
            return directories

    def filelist(self, timerange):
        """
        Returns the list of existent files in the archive for the given time
        range.

        Parameters
        ----------
        timerange : `~sunpy.time.TimeRange`
            Time interval where to find the directories for a given pattern.

        Returns
        -------
        filesurls : `list` of `str`
            List of all the files found between the time range given.

        Examples
        --------
        >>> from sunpy.net import Scraper
        >>> pattern = ('https://proba2.sidc.be/{instrument}/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/'
        ...            '{instrument}_lv1_{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}.fits')
        >>> swap = Scraper(pattern, instrument='swap')
        >>> from sunpy.time import TimeRange
        >>> timerange = TimeRange('2015-01-01T00:08:00','2015-01-01T00:12:00')
        >>> print(swap.filelist(timerange))  # doctest: +REMOTE_DATA
        ['https://proba2.sidc.be/swap/data/bsd/2015/01/01/swap_lv1_20150101_000857.fits',
         'https://proba2.sidc.be/swap/data/bsd/2015/01/01/swap_lv1_20150101_001027.fits',
         'https://proba2.sidc.be/swap/data/bsd/2015/01/01/swap_lv1_20150101_001157.fits']

        While writing the pattern, we can also leverage parse capabilities by using the ``{{}}`` notation to match parts of the filename that cannot be known beforehand:

        >>> from sunpy.net import Scraper
        >>> from sunpy.time import TimeRange
        >>> pattern = 'https://proba2.sidc.be/lyra/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/{{}}_lev{{Level:1d}}_std.fits'
        >>> lyra = Scraper(pattern)
        >>> timerange = TimeRange('2023-03-06T00:00:00','2023-03-06T00:10:00')
        >>> print(swap.filelist(timerange)) # doctest: +REMOTE_DATA
        ['https://proba2.sidc.be/swap/data/bsd/2023/03/06/swap_lv1_20230306_000128.fits',
        'https://proba2.sidc.be/swap/data/bsd/2023/03/06/swap_lv1_20230306_000318.fits',
        'https://proba2.sidc.be/swap/data/bsd/2023/03/06/swap_lv1_20230306_000508.fits',
        'https://proba2.sidc.be/swap/data/bsd/2023/03/06/swap_lv1_20230306_000658.fits',
        'https://proba2.sidc.be/swap/data/bsd/2023/03/06/swap_lv1_20230306_000848.fits']

        Notes
        -----
        The search is strict with the time range, so if the archive scraped contains daily files,
        but the range doesn't start from the beginning of the day, then the file for that day
        won't be selected. The end of the timerange will normally be OK as includes the file
        on such end time.
        """
        directories = self.range(timerange)
        if urlsplit(directories[0]).scheme == "ftp":
            return self._ftpfilelist(timerange)
        elif urlsplit(directories[0]).scheme == "file":
            return self._localfilelist(timerange)
        elif urlsplit(directories[0]).scheme in ["http", "https"]:
            return self._httpfilelist(timerange)
        else:
            return ValueError("The provided pattern should either be an FTP or a local file-path, or an HTTP address.")

    def _ftpfilelist(self, timerange):
        """
        Goes over archives available over ftp to return list of files in the given timerange.
        """
        directories = self.range(timerange)
        filesurls = list()
        ftpurl = urlsplit(directories[0]).netloc
        with FTP(ftpurl, user="anonymous", passwd="data@sunpy.org") as ftp:
            for directory in directories:
                try:
                    ftp.cwd(urlsplit(directory).path)
                except Exception as e:
                    log.debug(f"FTP CWD: {e}")
                    continue
                for file_i in ftp.nlst():
                    fullpath = directory + file_i
                    if self._url_follows_pattern(fullpath):
                        if self._check_timerange(fullpath, timerange):
                            filesurls.append(fullpath)

        filesurls = ['ftp://' + f"{urlsplit(url).netloc}{urlsplit(url).path}"
                     for url in filesurls]

        return filesurls

    def _localfilelist(self, timerange):
        """
        Goes over locally stored archives to return list of files in the given timerange.
        """
        pattern, datetime_pattern = self.pattern, self.datetime_pattern
        pattern_temp, datetime_pattern_temp = pattern.replace('file://', ''), datetime_pattern.replace('file://', '')
        if os.name == 'nt':
            pattern_temp = pattern_temp.replace('\\', '/')
            datetime_pattern_temp = datetime_pattern_temp.replace('\\', '/')
            prefix = 'file:///'
        else:
            prefix = 'file://'
        # Change pattern variables class-wide
        self.pattern, self.datetime_pattern = pattern_temp, datetime_pattern_temp
        directories = self.range(timerange)
        filepaths = list()
        for directory in directories:
            try:
                for file_i in os.listdir(directory):
                    fullpath = directory + file_i
                    if self._url_follows_pattern(fullpath):
                        if self._check_timerange(fullpath, timerange):
                            filepaths.append(fullpath)
            except FileNotFoundError:
                log.debug(f"Local directory not found: {directory}.")
        filepaths = [prefix + path for path in filepaths]
        # Set them back to their original values
        self.pattern, self.datetime_pattern = pattern, datetime_pattern
        return filepaths

    def _httpfilelist(self, timerange):
        """
        Goes over http archives hosted on the web, to return list of files in the given timerange.
        """
        directories = self.range(timerange)
        filesurls = list()
        retry_counts = {}
        while directories:
            directory = directories.pop(0)
            try:
                opn = urlopen(directory)
                try:
                    soup = BeautifulSoup(opn, "html.parser")
                    for link in soup.find_all("a"):
                        href = link.get("href")
                        if href is not None:
                            if href[0] == '/':
                                fullpath = self.domain + href[1:]
                            else:
                                fullpath = directory + href
                            if self._url_follows_pattern(fullpath):
                                if self._check_timerange(fullpath, timerange):
                                    filesurls.append(fullpath)
                finally:
                    opn.close()
            except HTTPError as http_err:
                # Ignore missing directories (issue #2684).
                if http_err.code == 404:
                    log.debug(f"Directory {directory} not found.")
                    continue
                if http_err.code in [400, 403]:
                    log.debug(f"Got error {http_err.code} while scraping {directory} : {http_err.reason}")
                    raise
                if http_err.code in [429, 504]:
                    # See if the server has told us how long to back off for
                    retry_after = http_err.hdrs.get('Retry-After', 2)
                    try:
                        # Ensure that we can parse the header as an int in sec
                        retry_after = int(retry_after)
                    except Exception as e:
                        log.debug(f"Converting retry_after failed: {e}")
                        retry_after = 2
                    log.debug(
                        f"Got {http_err.code} while scraping {directory}, waiting for {retry_after} seconds before retrying."
                    )
                    sleep(retry_after)
                    if retry_counts.get(directory, 0) > 4:
                        log.debug(f"Exceeded maximum retry limit for {directory}")
                        raise
                    retry_counts[directory] = retry_counts.get(directory, 0) + 1
                    directories.insert(0, directory)
                    continue
            except URLError as url_err:
                log.debug(f"Failed to parse content from {directory}: {url_err}")
                raise
            except Exception as e:
                log.debug(f"Failed to parse: {e}")
                raise
        return filesurls

    def _check_timerange(self, url, timerange):
        """
        Checks whether the time range represented in *url* intersects
        with the given time range.

        Parameters
        ----------
        url : `str`
            URL of the file.
        timerange : `~sunpy.time.TimeRange`
            Time interval for which files were searched.

        Returns
        -------
        `bool`
            `True` if URL's valid time range overlaps the given timerange, else `False`.
        """
        exdict = parse(self.pattern, url).named
        if exdict['year'] < 100:
            exdict['year'] = 2000 + exdict['year']
        if 'month' not in exdict:
            if 'month_name' in exdict:
                exdict['month'] = datetime.strptime(exdict['month_name'], '%B').month
            elif 'month_name_abbr' in exdict:
                exdict['month'] = datetime.strptime(exdict['month_name_abbr'], '%b').month
        tr = get_timerange_from_exdict(exdict)
        return tr.intersects(timerange)

    def _url_follows_pattern(self, url):
        """
        Check whether the url provided follows the pattern.
        """
        return parse(self.pattern, url)


    def _extract_date(self, url):
        """
        Extracts the date from a particular url following the pattern.
        """
        # Remove the user and password if present
        url = url.replace("anonymous:data@sunpy.org@", "")

        def url_to_list(txt):
            # Substitutes '.' and '_' for '/'.
            return re.sub(r'\.|_', '/', txt).split('/')

        # Create a list of all the blocks in times
        # assuming they are all separated with either '.', '_' or '/'
        pattern_list = url_to_list(self.pattern)
        url_list = url_to_list(url)
        time_order = [
            '%Y', '%y', '%b', '%B', '%m', '%d', '%j',
            '%H', '%I', '%M', '%S', '%e', '%f'
        ]
        final_date = []
        final_pattern = []
        # Find in directory and filename
        for pattern_elem, url_elem in zip(pattern_list, url_list):
            time_formats = [x for x in time_order if x in pattern_elem]
            if len(time_formats) > 0:
                # Find whether there's text that should not be here
                toremove = re.split('%.', pattern_elem)
                if len(toremove) > 0:
                    for bit in toremove:
                        if bit != '':
                            url_elem = url_elem.replace(bit, '', 1)
                            pattern_elem = pattern_elem.replace(bit, '', 1)
                final_date.append(url_elem)
                final_pattern.append(pattern_elem)
                for time_bit in time_formats:
                    time_order.remove(time_bit)
        # Find and remove repeated elements eg: %Y in ['%Y', '%Y%m%d']
        # Make all as single strings
        date_together = ''.join(final_date)
        pattern_together = ''.join(final_pattern)
        re_together = pattern_together
        for k, v in TIME_CONVERSIONS.items():
            re_together = re_together.replace(k, v)
        # Lists to contain the unique elements of the date and the pattern
        final_date = list()
        final_pattern = list()
        re_together = re_together.replace('[A-Z]', '\\[A-Z]')
        for p, r in zip(pattern_together.split('%')[1:], re_together.split('\\')[1:]):
            if p == 'e':
                continue
            regexp = fr'\{r}' if not r.startswith('[') else r
            pattern = f'%{p}'
            date_part = re.search(regexp, date_together)
            date_together = date_together[:date_part.start()] \
                + date_together[date_part.end():]
            if pattern not in final_pattern:
                final_pattern.append(f'%{p}')
                final_date.append(date_part.group())
        return Time.strptime(' '.join(final_date),
                             ' '.join(final_pattern))

    def _extract_files_meta(self, timerange, matcher=None):
        """
        Returns metadata information contained in URLs.

        Parameters
        ----------
        timerange : `~sunpy.time.TimeRange`
            Time interval where to find the directories for a given pattern.
        matcher : `dict`, optional
            Dictionary to check if extracted metadata is valid.
            Defaults to None.

        Returns
        -------
        `list` of `dict`
            List of metadata info for all URLs.
        """
        urls = self.filelist(timerange)
        metalist = []
        for url in urls:
            metadict = parse(self.pattern, url)
            if metadict is not None:
                append = True
                metadict = metadict.named
                metadict['url'] = url
                if 'month' not in metadict:
                    if 'month_name' in metadict:
                        metadict['month'] = datetime.strptime(metadict['month_name'], '%B').month
                    elif 'month_name_abbr' in metadict:
                        metadict['month'] = datetime.strptime(metadict['month_name_abbr'], '%b').month
                if matcher is not None:
                    for k in metadict:
                        if match := matcher.get(k):
                            if str(metadict[k]) not in match:
                                append = False
                                break
                if append:
                    metalist.append(metadict)
        return metalist



"""Tests for PSP/WISPR"""
from textwrap import dedent

import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits

from sunpy.map import Map
from sunpy.map.sources import WISPRMap
from .helpers import _test_private_date_setters


@pytest.fixture
def wispr_map(scope="module"):
    # Please do not edit this header, it is a 1:1 copy of a SOLO EUI FSI 304 map
    raw_header = dedent("""\
        SIMPLE  =                    T / Written by IDL:  Wed Sep  9 14:48:00 2020
        BITPIX  =                   32 /  32-bit twos complement binary integer
        NAXIS   =                    2 /
        NAXIS1  =                  960 /
        NAXIS2  =                 1024 /
        FILENAME= 'psp_L1_wispr_20200125T000229_V1_2302.fits' /
        FILE_RAW= 'CmpB.12EE4DE3'      /
        APID    = '41b     '           /
        DATE    = '2020-09-09T18:47:59.761' /
        DATE-OBS= '2020-01-25T00:02:29.618' /
        DATE-BEG= '2020-01-25T00:02:29.618' /
        DATE-AVG= '2020-01-25T00:08:20.842' /
        DATE-END= '2020-01-25T00:14:12.067' /
        OBT_BEG =        317606692.959 /
        OBT_END =        317607252.950 /
        NSUMEXP =                    5 /
        NSUMBAD =                    0 /
        XPOSURE =              700.000 /
        IP_TIMET=                  139 /
        READTIME=              2.45760 /
        TELAPSE =              702.449 /
        TIMESYS = 'UTC     '           /
        LEVEL   = 'L1      '           /
        CREATOR = 'wispr_reduce.pro,v 1.63' /
        ORIGIN  = 'NRL     '           /
        DETECTOR=                    2 /
        REGION  =                    2 /
        READOUT0=                    3 /
        CAMERA  = 'FMCFMD  '           /
        CICUCODE= '129     '           /
        DRBUCODE= '130     '           /
        DRB1UCOD= '130     '           /
        DRB2UCOD= '130     '           /
        OBJECT  = 'OuterFFV'           /
        OBS_MODE= 'SYNOPTIC'           /
        TARGET  = 'VexpOuterL.os'      /
        TIMELINE= '584,0x1a050'        /
        STUDY_ID=                  742 /
        STUDYCNT=                    1 /
        VERS_CAL= '5; 57   '           /
        PRIORITY=                    1 /
        VERSION =                    1 /
        FINAL   =                    T /
        BSCALE  =              1.00000 /
        BZERO   =              0.00000 /
        BUNIT   = 'DN      '           /
        DATAMIN =              856.000 /  Minimum Value Not Equal to Zero before BSCALE
        DATAMAX =              249069. /  Maximum Value before BSCALE
        BLANK   =                61166 /
        PXBEG1  =                    1 /
        PXEND1  =                 2048 /
        PXBEG2  =                    1 /
        PXEND2  =                 1920 /
        R1COL   =                    1 /
        R2COL   =                 1920 /
        R1ROW   =                    1 /
        R2ROW   =                 2048 /
        RECTIFY =                    T /
        RECTROTA=                    6 /
        NBIN1   =                    2 /
        NBIN2   =                    2 /
        NBIN    =              4.00000 /
        COMPRESS= 'Lossless'           /
        COMP_RAT=             0.675516 /
        COSMICR =                    T /
        DSTART1 =                    1 /
        DSTOP1  =                  955 /
        DSTART2 =                    1 /
        DSTOP2  =                 1019 /
        IMGCTR  =                 8151 /
        CCIMGSEQ=                 2505 /
        CCEXPCTL=                  999 /
        CCIMGCTR=                   15 /
        COSMICS =                35210 /
        CAMSATPX=                    0 /
        CAMSTRVX=                18638 /
        IP_INIT =                    2 /
        IPIMBUFF=                30840 /
        IPCMBUFF=                  320 /
        IPSATPX =                16000 /
        IPSTRVPX=                  300 /
        IPMINPX =                    0 /
        IPMAXPX =                16383 /
        IPTRUNC =                    0 /
        IPBIAS  =                    2 /
        IPCRSMUL=              4.00000 /
        IPBITSPP=                   19 /
        IPMASK  =                    0 /
        IPMASKCR= '144,879,0,959'      /
        LEDSTATE= 'Off     '           /
        LEDDAC  =                    0 /
        OFFSET  =                 2256 /
        GAINCMD =                   12 /
        GAINMODE= 'HIGH    '           /
        IP_FUNC =                16584 /
        IPCMPCTL=                  305 /
        XFBYTES =              1577140 /
        SCFLAGS = '0x0015ff3e'         /
        CIE_T   =             -39.9500 /
        DRB1_T  =             -39.3083 /
        DET1_T  =             -70.9754 /
        DRB2_T  =             -39.0940 /
        DET2_T  =             -70.9950 /
        ISPREG0 =                   28 /
        ISPREG1 =                61166 /
        ISPREG2 =                61166 /
        ISPREG3 =                61166 /
        ISPREG4 =                61166 /
        ISPREG5 =                61166 /
        ISPREG6 =                  825 /
        ISPREG7 =                  742 /
        UCODREG0=                61166 /
        UCODREG1=                  128 /
        UCODREG2=                    0 /
        UCODREG3=                    0 /
        UCODREG4=                 1920 /
        UCODREG5=                61166 /
        UCODREG6=                61166 /
        UCODREG7=                61166 /
        DSUN_OBS=        31678123000.0 /
        SC_YAW  =            0.0127380 /
        SC_PITCH=         -0.000853018 /
        SC_ROLL =             -3.83791 /
        DATAZER =                    0 /  Number of Zero Pixels
        DATASAT =                    0 /  Number of Saturated Pixels
        DSATVAL =               524287 /  Value used as saturated
        DATAAVG =              107440. /  Between 0 and DSATVAL
        DATAMDN =              96257.0 /  Between 0 and DSATVAL
        DATASIG =              43882.8 /  Between 0 and DSATVAL
        DATAP01 =                 5409 /
        DATAP10 =                63754 /
        DATAP25 =                77544 /
        DATAP50 =                96257 /
        DATAP75 =               128137 /
        DATAP90 =               171321 /
        DATAP95 =               200730 /
        DATAP98 =               230366 /
        DATAP99 =               237914 /
        OBSRVTRY= 'Parker Solar Probe' /
        INSTRUME= 'WISPR   '           /
        WCSNAME = 'Helioprojective Zenith Polynomial' /
        CTYPE1  = 'HPLN-ZPN'           /
        CTYPE2  = 'HPLT-ZPN'           /
        CUNIT1  = 'deg     '           /
        CUNIT2  = 'deg     '           /
        CRPIX1  =              492.333 /
        CRPIX2  =              513.174 /
        PC1_1   =       0.998328091852 /
        PC1_2   =      0.0578015658869 /
        PC2_1   =     -0.0578015658869 /
        PC2_2   =       0.998328091852 /
        CDELT1  =            0.0570821 /
        CDELT2  =            0.0570821 /
        CRVAL1  =        76.6509814623 /
        CRVAL2  =       -13.2570275339 /
        ATT_FILE= 'spp_2020_025_02.ah.bc' /
        PV1_1   =        0.00000000000 /
        PV1_2   =        90.0000000000 /
        PV1_3   =        180.000000000 /
        PV2_0   =   -0.000429731007898 /
        PV2_1   =        1.01042997837 /
        PV2_2   =     -0.0547256991267 /
        PV2_3   =      0.0849808976054 /
        PV2_4   =      -0.175265997648 /
        PV2_5   =      0.0547404997051 /
        LONPOLE =        180.000000000 /
        LATPOLE =        0.00000000000 /
        CTYPE1A = 'RA---ZPN'           /
        CTYPE2A = 'DEC--ZPN'           /
        CUNIT1A = 'deg     '           /
        CUNIT2A = 'deg     '           /
        PC1_1A  =       0.932533562542 /
        PC1_2A  =      -0.361083307193 /
        PC2_1A  =       0.361083307193 /
        PC2_2A  =       0.932533562542 /
        CRPIX1A =              492.333 /
        CRPIX2A =              513.174 /
        CRVAL1A =        170.519696093 /
        CRVAL2A =       -2.59678257890 /
        CDELT1A =           -0.0570821 /
        CDELT2A =            0.0570821 /
        PV1_1A  =        0.00000000000 /
        PV1_2A  =        90.0000000000 /
        PV1_3A  =        180.000000000 /
        PV2_0A  =   -0.000429731007898 /
        PV2_1A  =        1.01042997837 /
        PV2_2A  =     -0.0547256991267 /
        PV2_3A  =      0.0849808976054 /
        PV2_4A  =      -0.175265997648 /
        PV2_5A  =      0.0547404997051 /
        LONPOLEA=        180.000000000 /
        LATPOLEA=        0.00000000000 /
        RSUN_ARC=        4524.67854698 /
        RSUN_REF=        695507968.000 /
        SOLAR_EP=        15.9285776205 /
        CAR_ROT =                 2226 /
        HGLT_OBS=       0.390715276104 /
        HGLN_OBS=       -54.7633009507 /
        CRLT_OBS=       0.390715276104 /
        CRLN_OBS=        59.8905844055 /
        HEEX_OBS=        18187298676.0 /
        HEEY_OBS=       -25963488915.7 /
        HEEZ_OBS=       -241652330.942 /
        HCIX_OBS=        31492170156.9 /
        HCIY_OBS=       -3624289322.90 /
        HCIZ_OBS=        216174389.143 /
        HCIX_VOB=       -38361.9314129 /
        HCIY_VOB=        71493.5202110 /
        HCIZ_VOB=       -4786.08742968 /
        HAEX_OBS=        11254842547.2 /
        HAEY_OBS=        29634581727.6 /
        HAEZ_OBS=       -243044691.612 /
        HEQX_OBS=        18289512873.5 /
        HEQY_OBS=       -25891812859.2 /
        HEQZ_OBS=        216174389.143 /
        EAR_TIME=        383.108719220 /
        SUN_TIME=        105.742395411 /
        HISTORY make_wispr_hdr.pro,v 1.57
        HISTORY wincs-gse.nrl.navy.mil
        HISTORY FSW version 932
        HISTORY wispr_reduce.pro,v 1.63
        HISTORY wispr_readrawsci.pro,v 1.17
        HISTORY Id: get_wispr_pointing.pro,v 1.29 2020/06/02 19:29:24 nathan Exp
        """)
    header = fits.Header.fromstring(raw_header, sep='\n')
    data = np.random.rand(header['naxis1'], header['naxis2'])
    return Map(data, header)


def test_WISPRMap(wispr_map):
    assert isinstance(wispr_map, WISPRMap)


def test_is_datasource_for(wispr_map):
    assert wispr_map.is_datasource_for(wispr_map.data, wispr_map.meta)


def test_observer_coordinate(wispr_map):
    obs_coord = wispr_map.observer_coordinate
    assert isinstance(obs_coord, SkyCoord)
    assert obs_coord.obstime.isot == wispr_map.meta['date-avg']


def test_observatory(wispr_map):
    assert wispr_map.observatory == "Parker Solar Probe"


def test_reference_date(wispr_map):
    assert wispr_map.reference_date.isot == "2020-01-25T00:08:20.842"


def test_date(wispr_map):
    assert wispr_map.date.isot == "2020-01-25T00:02:29.618"


def test_private_date_setters(wispr_map):
    _test_private_date_setters(wispr_map)


def test_measurement(wispr_map):
    assert wispr_map.measurement is None


def test_wavelength(wispr_map):
    assert wispr_map.wavelength is None


def test_exposure_time(wispr_map):
    assert wispr_map.exposure_time == u.Quantity(700, 's')


def test_processing_level(wispr_map):
    assert wispr_map.processing_level == 1

    for value, expected in [
            ('L1', 1),
            ('L2', 2),
            ('L2b', '2b'),
            ('L3', 3),
            ('LW', 'W')]:
        wispr_map.meta['level'] = value
        assert wispr_map.processing_level == expected


def test_detector(wispr_map):
    assert wispr_map.detector == 'Outer'

    wispr_map.meta['DETECTOR'] = 1
    assert wispr_map.detector == 'Inner'

    wispr_map.meta['DETECTOR'] = 'other_val'
    assert wispr_map.detector == 'other_val'


def test_unit(wispr_map):
    assert wispr_map.unit == u.Unit('DN')


def test_norm_clip(wispr_map):
    # Tests that the default normalizer has clipping disabled
    assert not wispr_map.plot_settings['norm'].clip


def test_name(wispr_map):
    assert wispr_map.name == 'WISPR Outer 2020-01-25 00:02:29'


def test_wcs(wispr_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    wispr_map.pixel_to_world(0*u.pix, 0*u.pix)



"""TRACE Map subclass definitions"""

import astropy.units as u
from astropy.visualization import LogStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.map.mapbase import GenericMap, SpatialPair
from sunpy.map.sources.source_type import source_stretch

__all__ = ['TRACEMap']


class TRACEMap(GenericMap):
    """TRACE Image Map

    The Transition Region and Coronal Explorer was a
    NASA Small Explorer (SMEX) mission to image the
    solar corona and transition region at high angular and temporal resolution.
    TRACE observed the Sun in the following passbands, 5000 A, 1700 A, 1600 A,
    1550 A (C IV), 1216 A (H1 Lyman-alpha), 173 A (Fe IX), 195 A (Fe XII),
    and 284 A (Fe XV). TRACE provides solar images with an 8.5 x 8.5 arcminute
    field of view and 0.5 arcsecond pixels. It was placed in a sun-synchronous
    orbit, enabling it to make continuous solar observations.

    The TRACE mission operated was launched on 2 April 1998 and obtained its
    last science image on 6 June 2010 23:56 UT.

    References
    ----------
    * `Mission/Instrument Page <https://sdowww.lmsal.com/TRACE>`__
    * `Fits headers <https://sdowww.lmsal.com/TRACE/Project/Instrument/cal/>`__
    * `Analysis Guide <https://sdowww.lmsal.com/TRACE/tag/>`__
    * `Passband reference <https://sdowww.lmsal.com/TRACE/Project/Instrument/inspass.htm>`__

    .. note::

        Note that this map definition is currently only being tested on JPEG2000
        files. TRACE FITS data is stored in a more complex format. Typically
        TRACE data is stored in hourly "tri" files that store all the data taken
        by TRACE in the hour indicated by the filename. Those files must first be
        understood and parsed to obtain the science data. The ability to do this
        is not yet in SunPy, but is available in SSWIDL. Please refer to the links
        above concerning how to read "tri" files in SSWIDL.
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self._nickname = self.detector
        # Colour maps
        self.plot_settings['cmap'] = 'trace' + str(self.meta['WAVE_LEN'])
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, LogStretch()), clip=False)

    @property
    def coordinate_system(self):
        """
        Override the default implementation to handle TRACEMAP-specific logic for CTYPE values.
        """
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1.lower() in ("solar-x", "solar_x"):
            ctype1 = 'HPLN-TAN'
        if ctype2.lower() in ("solar-y", "solar_y"):
            ctype2 = 'HPLT-TAN'
        return SpatialPair(ctype1, ctype2)

    @property
    def spatial_units(self):
        """
        If not present in CUNIT{1,2} keywords, defaults to arcsec.
        """
        return SpatialPair(u.Unit(self.meta.get('cunit1', 'arcsec')),
                           u.Unit(self.meta.get('cunit2', 'arcsec')))

    @property
    def observatory(self):
        return "TRACE"

    @property
    def detector(self):
        return "TRACE"

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an TRACE image"""
        return header.get('instrume') == 'TRACE'

    @property
    def measurement(self):
        s = self.meta['WAVE_LEN']
        if s == 'WL':
            s = 'white-light'
        return s



import numpy as np
import pytest

spiceypy = pytest.importorskip("spiceypy")

import astropy.units as u
from astropy.coordinates import ConvertError, SkyCoord, UnitSphericalRepresentation, frame_transform_graph
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import download_file

from sunpy.coordinates import Helioprojective, get_earth, spice
from sunpy.time import parse_time


@pytest.fixture(scope="module", params=[pytest.param(None, marks=pytest.mark.remote_data)])
def spice_test():
    # We use the cache for the SPK file because it's also used in other tests (does cache work?)
    spk = download_file("https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/de440s.bsp",
                        cache=True)

    # Implicitly test adding a single kernel, but no kernel frames are installed here
    spice.initialize(spk)

    # Leapseconds (LSK) and physical constants (PCK)
    lsk = download_file("https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls")
    pck = download_file("https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/pck00011_n0066.tpc")

    # Use frames (FK) from SunSPICE
    fk = download_file("https://soho.nascom.nasa.gov/solarsoft/packages/sunspice/data/heliospheric.tf")

    # Use more frames (FK and IK) from Solar Orbiter
    solo_sc_fk = download_file("https://spiftp.esac.esa.int/data/SPICE/SOLAR-ORBITER/kernels/fk/solo_ANC_soc-sc-fk_V09.tf")
    solo_spd_ik = download_file("https://spiftp.esac.esa.int/data/SPICE/SOLAR-ORBITER/kernels/ik/solo_ANC_soc-epd-ik_V03.ti")

    # Implicitly test adding a kernel list
    spice.initialize([lsk, pck, fk, solo_sc_fk, solo_spd_ik])

    yield

    # The fact that we need to do this suggests that the cache is not persistent across tests...
    spiceypy.unload(spk)


def test_frame_creation(spice_test):
    expected = {"spice_ECLIPDATE", "spice_GSE", "spice_HCI", "spice_HEE", "spice_HEEQ", "spice_GEORTN"}
    assert expected.issubset(frame_transform_graph.get_names())


def test_double_initialize(spice_test):
    # Smoke test for an issue with calling initialize() more than once a session
    spice.initialize([])

    expected = {"spice_ECLIPDATE", "spice_GSE", "spice_HCI", "spice_HEE", "spice_HEEQ", "spice_GEORTN"}
    assert expected.issubset(frame_transform_graph.get_names())


def test_install_frame(spice_test):
    # Installing by ID is already tested through a successful initialize()

    assert "spice_IAU_SUN" not in frame_transform_graph.get_names()
    spice.install_frame('IAU_SUN')
    assert "spice_IAU_SUN" in frame_transform_graph.get_names()

    assert "spice_IAU_EARTH" not in frame_transform_graph.get_names()
    spice.install_frame('iau_earth')
    assert "spice_IAU_EARTH" in frame_transform_graph.get_names()


def test_install_frame_bad_id(spice_test):
    with pytest.raises(ValueError, match="not a valid SPICE frame ID"):
        spice.install_frame(999999)


def test_install_frame_bad_name(spice_test):
    with pytest.raises(ValueError, match="not a valid SPICE frame name"):
        spice.install_frame('NOT_A_FRAME_NAME')


def test_transformation(spice_test):
    coord = SkyCoord(1e7*u.km, 1e6*u.km, 1e5*u.km, representation_type='cartesian',
                     frame='spice_GSE', obstime='2023-10-17')
    new_coord = coord.spice_HEE
    new_coord.representation_type ='cartesian'

    assert_quantity_allclose(new_coord.x, 1.49128669e8*u.km - coord.x)
    assert_quantity_allclose(new_coord.y, -coord.y)
    assert_quantity_allclose(new_coord.z, coord.z)


def test_transformation_array_time(spice_test):
    coord = SkyCoord([1e7, 1e7]*u.km, [1e6, 1e6]*u.km, [1e5, 1e5]*u.km,
                     representation_type='cartesian',
                     frame='spice_GSE', obstime=['2023-10-17', '2023-10-18'])
    new_coord = coord.spice_HEE
    new_coord.representation_type ='cartesian'

    assert len(new_coord) == 2
    assert_quantity_allclose(new_coord.x, [1.49128669e8, 1.49085802e+08]*u.km - coord.x)
    assert_quantity_allclose(new_coord.y, -coord.y)
    assert_quantity_allclose(new_coord.z, coord.z)


def test_transformation_2d_okay(spice_test):
    coord = SkyCoord(30*u.deg, 45*u.deg, frame='spice_HCI', obstime='2023-10-17')
    new_coord = coord.spice_HEE

    assert isinstance(new_coord.data, UnitSphericalRepresentation)


def test_transformation_2d_not_okay(spice_test):
    coord = SkyCoord(30*u.deg, 45*u.deg, frame='spice_HCI', obstime='2023-10-17')
    with pytest.raises(ConvertError, match="due to a shift in origin"):
        assert coord.icrs

    coord = SkyCoord(30*u.deg, 45*u.deg, frame='icrs', obstime='2023-10-17')
    with pytest.raises(ConvertError, match="due to a shift in origin"):
        assert coord.spice_HCI


def test_get_body(spice_test):
    # Regression test
    earth_by_id = spice.get_body(399, '2023-10-17')
    earth_by_name = spice.get_body('earth', '2023-10-17')

    assert earth_by_id.name == 'icrs'
    assert_quantity_allclose(earth_by_id.ra, 21.3678774*u.deg)
    assert_quantity_allclose(earth_by_id.dec, 8.9883346*u.deg)
    assert_quantity_allclose(earth_by_id.distance, 1.47846987e8*u.km)

    assert earth_by_name == earth_by_id


def test_get_body_observer(spice_test):
    # Regression test
    earth = spice.get_body('earth', '2023-10-17')
    venus = spice.get_body('venus', '2023-10-17')
    venus_earth = spice.get_body('venus', '2023-10-17', observer=earth)
    venus_earth_hci = spice.get_body('venus', '2023-10-17', observer=earth, spice_frame='HCI')

    assert_quantity_allclose(venus_earth.separation_3d(venus), 11237.0641158*u.km)
    assert_quantity_allclose(venus_earth_hci.separation_3d(venus), 11237.0641158*u.km)


def test_get_body_array_time(spice_test):
    # Regression test
    obstime = parse_time(['2013-10-17', '2013-10-18'])
    earth = spice.get_body(399, obstime, spice_frame='HCI')

    assert len(earth) == 2
    assert earth[0].obstime == obstime[0]
    assert earth[1].obstime == obstime[1]
    assert earth[0].lon != earth[1].lon
    assert earth[0].lat != earth[1].lat
    assert earth[0].distance != earth[1].distance


def test_get_body_spice_frame(spice_test):
    # Regression test
    moon_gse = spice.get_body('moon', '2023-10-17', spice_frame='GSE')
    moon_gse.representation_type = 'cartesian'
    moon_hee = spice.get_body('moon', '2023-10-17', spice_frame='HEE')
    moon_hee.representation_type = 'cartesian'

    assert moon_gse.name == 'spice_GSE'
    assert_quantity_allclose(moon_gse.x, 349769.2488428*u.km)
    assert_quantity_allclose(moon_gse.y, 171312.08978192*u.km)
    assert_quantity_allclose(moon_gse.z, -14990.88338588*u.km)

    assert moon_hee.name == 'spice_HEE'
    assert_quantity_allclose(moon_hee.x, 1.487789e+08*u.km)
    assert_quantity_allclose(moon_hee.y, -171312.08978202*u.km)
    assert_quantity_allclose(moon_hee.z, -14990.88338588*u.km)


def test_get_fov_rectangle(spice_test):
    fov = spice.get_fov('SOLO_EPD_STEP', '2023-10-17')
    width = 27*u.deg
    height = 13*u.deg
    # Because SPICE defines the rectangular FOV on a plane,
    # the height in spherical latitude is shorter at the corners
    adjusted_height = np.arctan(np.tan(height) * np.cos(width))
    assert_quantity_allclose(fov.lon, [1, -1, -1, 1] * width)
    assert_quantity_allclose(fov.lat, [1, 1, -1, -1] * adjusted_height)


def test_get_fov_circle(spice_test):
    obstime = parse_time(['2023-10-17', '2023-10-18'])
    fov = spice.get_fov('SOLO_EPD_SIS_ASW', obstime)
    angle = 11*u.deg
    assert fov.shape == (2, 100)
    assert (fov.obstime[:, 0] == obstime).all()
    assert_quantity_allclose(fov[0].cartesian.xyz, fov[1].cartesian.xyz)
    assert_quantity_allclose(fov[0, [0, 25, 50, 75]].lon, [1, 0, -1, 0] * angle, atol=1e-10*u.deg)
    assert_quantity_allclose(fov[0, [0, 25, 50, 75]].lat, [0, -1, 0, 1] * angle, atol=1e-10*u.deg)


def test_to_helioprojective(spice_test):
    # Define a coordinate in SunSPICE's GSE frame
    spice_coord = SkyCoord(UnitSphericalRepresentation(0.1*u.deg, 0.2*u.deg),
                           frame='spice_GSE', obstime='2023-10-17')
    hpc = spice_coord.to_helioprojective()

    assert isinstance(hpc.frame, Helioprojective)

    # The origin of spice_GSE (the Earth) should be used as the Helioprojective observer
    # We loosen tolerances because the planetary ephemeris is different
    earth = get_earth(hpc.obstime)
    assert_quantity_allclose(hpc.observer.lon, earth.lon, atol=1e-5*u.deg)
    assert_quantity_allclose(hpc.observer.lat, earth.lat, rtol=1e-6)
    assert_quantity_allclose(hpc.observer.radius, earth.radius, rtol=1e-6)

    # The angular offset from disk center should exactly match
    assert_quantity_allclose(np.arccos(hpc.cartesian.x), 0.22360671*u.deg)

    # We test the correctness of the HPC coordinate by transforming back to *our* GSE frame
    # and comparing against the original values in the SunSPICE GSE frame
    sunpy_coord = hpc.geocentricsolarecliptic
    assert_quantity_allclose(sunpy_coord.lon, spice_coord.lon, rtol=1e-6)
    assert_quantity_allclose(sunpy_coord.lat, spice_coord.lat, rtol=1e-6)

def test_get_rotation_matrix(spice_test):
    result1 = spice.get_rotation_matrix('HEEQ', 'J2000', '2024-07-04')
    expected_result1 = np.array([[ 0.20454304,  0.97118061,  0.12235349],
                                 [-0.87442558,  0.23746563, -0.42307208],
                                 [-0.43993415, -0.02045257,  0.8977971]])
    np.testing.assert_allclose(result1, expected_result1, atol=1e-6)

    result2 = spice.get_rotation_matrix('HEEQ', 'HEEQ', '2024-07-04', '2024-07-18')
    expected_result2 = np.array([[ 0.97314148,  0.23020788,  0],
                                 [-0.23020788,  0.97314148, 0],
                                 [0, 0, 1]])
    np.testing.assert_allclose(result2, expected_result2, atol=1e-6)



"""
====================================
Drawing a rotated rectangle on a map
====================================

This example will demonstrate how to draw a rectangle that is rotated relative
to the axes on a map using :meth:`~sunpy.map.GenericMap.draw_quadrangle`.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord, SkyOffsetFrame

import sunpy.data.sample
import sunpy.map

################################################################################
# Let's start with a sample image of AIA 171.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)

################################################################################
# Define the rotation angle and center coordinate of the rectangle,
# as well as the width and height in physical units.

rotation_angle = 30 * u.deg
center_coord = SkyCoord(-100 * u.arcsec, -300 * u.arcsec, frame=aia_map.coordinate_frame)
width = 1000 * u.arcsec
height = 300 * u.arcsec

################################################################################
# We define a custom coordinate frame for the rotated rectangle by providing
# the center in the AIA 171 coordinate frame and rotation angle to
# `~astropy.coordinates.SkyOffsetFrame`. We then define a 2-element
# `~astropy.coordinates.SkyCoord` in that custom coordinate frame for the
# bottom-left and top-right corners of the rectangle.

offset_frame = SkyOffsetFrame(origin=center_coord, rotation=rotation_angle)
rectangle = SkyCoord(lon=[-1/2, 1/2] * width, lat=[-1/2, 1/2] * height, frame=offset_frame)

################################################################################
# Finally, we will draw the rotated rectangle and its center.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_map)
aia_map.plot(axes=ax, clip_interval=(1, 99.99) * u.percent)
ax.plot_coord(center_coord, "o", color="red")

aia_map.draw_quadrangle(
    rectangle,
    axes=ax,
    edgecolor="red",
    linestyle="--",
    linewidth=2,
)

plt.show()



import calendar
from datetime import datetime, timedelta

from dateutil.relativedelta import relativedelta

from sunpy.time import TimeRange

__all__ = ["extract_timestep", "date_floor", "get_timerange_from_exdict"]

TIME_QUANTITIES = {
    'day': timedelta(days=1),
    'hour': timedelta(hours=1),
    'minute': timedelta(minutes=1),
    'second': timedelta(seconds=1),
    'millisecond': timedelta(milliseconds=1)
}


def extract_timestep(directoryPattern):
    """
    Obtain the smallest time step for the given pattern.

    Parameters
    ----------
    directoryPattern : `str`
        The pattern containing the datetime-formats

    Returns
    -------
    `dateutil.relativedelta.relativedelta`
        The smallest timestep unit from the provided pattern.
    """
    if "%S" in directoryPattern:
        return relativedelta(seconds=1)
    elif "%M" in directoryPattern:
        return relativedelta(minutes=1)
    elif any(hour in directoryPattern for hour in ["%H"]):
        return relativedelta(hours=1)
    elif any(day in directoryPattern for day in ["%d", "%j"]):
        return relativedelta(days=1)
    elif any(month in directoryPattern for month in ["%b", "%B", "%m"]):
        return relativedelta(months=1)
    elif any(year in directoryPattern for year in ["%Y", "%y"]):
        return relativedelta(years=1)
    else:
        return None


def date_floor(date, timestep):
    """
    Return the "floor" of the given date and timestep.

    Parameters
    ----------
    datetime : `datetime.datetime` or `astropy.time.Time`
        The date to floor.
    timestep : `dateutil.relativedelta.relativedelta`
        The smallest timestep to floor.

    Returns
    -------
    `datetime.datetime`
        The time floored at the given time step
    """
    date_parts = [int(p) for p in date.strftime('%Y,%m,%d,%H,%M,%S').split(',')]
    date_parts[-1] = date_parts[-1] % 60
    date = datetime(*date_parts)
    orig_time_tup = date.timetuple()
    time_tup = [orig_time_tup.tm_year, orig_time_tup.tm_mon, orig_time_tup.tm_mday,
                orig_time_tup.tm_hour, orig_time_tup.tm_min, orig_time_tup.tm_sec]
    if timestep == relativedelta(minutes=1):
        time_tup[-1] = 0
    elif timestep == relativedelta(hours=1):
        time_tup[-2:] = [0, 0]
    elif timestep == relativedelta(days=1):
        time_tup[-3:] = [0, 0, 0]
    elif timestep == relativedelta(months=1):
        time_tup[-4:] = [1, 0, 0, 0]
    elif timestep == relativedelta(years=1):
        time_tup[-5:] = [1, 1, 0, 0, 0]

    return datetime(*time_tup)


def get_timerange_from_exdict(exdict):
    """
    Function to get URL's timerange using extracted metadata.
    It computes start and end times first using the given
    dictionary and then returns a timerange.

    Parameters
    ----------
    exdict : `dict`
        Metadata extracted from the file's url.

    Returns
    -------
    `~sunpy.time.TimeRange`
        The time range of the file.
    """
    # This function deliberately does NOT use astropy.time because it is not
    # needed, and the performance overheads in dealing with astropy.time.Time
    # objects are large
    datetypes = ['year', 'month', 'day']
    timetypes = ['hour', 'minute', 'second', 'millisecond']
    dtlist = [int(exdict.get(d, 1)) for d in datetypes]
    dtlist.extend([int(exdict.get(t, 0)) for t in timetypes])
    startTime = datetime(*dtlist)
    tdelta = TIME_QUANTITIES['millisecond']
    if "second" in exdict:
        tdelta = TIME_QUANTITIES['second']
    elif "minute" in exdict:
        tdelta = TIME_QUANTITIES['minute']
    elif "hour" in exdict:
        tdelta = TIME_QUANTITIES['hour']
    elif "day" in exdict:
        tdelta = TIME_QUANTITIES['day']
    elif "month" in exdict:
        days_in_month = calendar.monthrange(int(exdict['year']), int(exdict['month']))[1]
        tdelta = days_in_month*TIME_QUANTITIES['day']
    elif "year" in exdict:
        if calendar.isleap(int(exdict['year'])):
            tdelta = 366*TIME_QUANTITIES['day']
        else:
            tdelta = 365*TIME_QUANTITIES['day']
    endTime = startTime + tdelta - TIME_QUANTITIES['millisecond']
    return TimeRange(startTime, endTime)



import pytest

import sunpy.net.dataretriever.sources.gong as gong
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.net._attrs import Instrument, Time
from sunpy.net.dataretriever.client import QueryResponse
from sunpy.net.fido_factory import UnifiedResponse


@pytest.fixture
def GSClient():
    return gong.GONGClient()


@pytest.mark.remote_data
@pytest.mark.parametrize(("timerange", "url_start", "url_end"), [
    (a.Time('2020/1/30', '2020/2/1'),
     'https://gong2.nso.edu/oQR/zqs/202001/mrzqs200130/mrzqs200130t0004c2227_349.fits.gz',
     'https://gong2.nso.edu/oQR/zqs/202001/mrzqs200131/mrzqs200131t2314c2227_323.fits.gz'
     ),
    (a.Time('2020/4/21', '2020/4/22'),
     'https://gong2.nso.edu/oQR/zqs/202004/mrzqs200421/mrzqs200421t0004c2230_348.fits.gz',
     'https://gong2.nso.edu/oQR/zqs/202004/mrzqs200421/mrzqs200421t2314c2230_335.fits.gz'
     ),
    (a.Time('2006/9/19', '2006/9/19 22:00'),
     'https://gong2.nso.edu/oQR/zqs/200609/mrzqs060919/mrzqs060919t1154c2048_323.fits.gz',
     'https://gong2.nso.edu/oQR/zqs/200609/mrzqs060919/mrzqs060919t1754c2048_320.fits.gz'
     )
])
def test_get_url_for_time_range(GSClient, timerange, url_start, url_end):
    qresponse = GSClient.search(timerange)
    urls = [i['url'] for i in qresponse]
    assert isinstance(urls, list)
    assert urls[0] == url_start
    assert urls[-1] == url_end


@pytest.mark.parametrize(
    ("query", "result"),
    [((a.Time('2020/1/1', '2020/1/2'), a.Instrument('gong')), True),
     ((a.Time('2020/1/1', '2020/1/2'), a.Instrument('goes')), False),
     ((a.Time('2020/1/1', '2020/1/2'), a.Instrument('gong'), a.Physobs('LOS_MAGNETIC_FIELD'),
       a.ExtentType('synoptic')), True),
     ((a.Time('2020/1/1', '2020/1/2'), a.Instrument('gong'),
       a.ExtentType('synoptic')), True),
     ((a.Time('2020/1/1', '2020/1/2'), a.Instrument('gong'),
       a.ExtentType('FULL_DISK')), False)])
def test_can_handle_query(query, result):
    assert gong.GONGClient._can_handle_query(*query) == result


@pytest.mark.remote_data
@pytest.mark.parametrize(("time", "instrument"), [
    (Time('2013/8/27', '2013/8/27'), Instrument('gong')),
    (Time('2020/4/23 17:00', '2020/4/23 21:00'), Instrument('gong')),
])
def test_get(GSClient, time, instrument):
    qr1 = GSClient.search(time, instrument)
    assert isinstance(qr1, QueryResponse)
    download_list = GSClient.fetch(qr1)
    assert len(download_list) == len(qr1)


@pytest.mark.remote_data
@pytest.mark.parametrize(
    ("time", "instrument"),
    [(a.Time('2019/10/4', '2019/10/4 2:00'), a.Instrument('gong')),
     (a.Time('2019/12/31 21:00', '2020/1/1'), a.Instrument('gong'))])
def test_fido(time, instrument):
    qr = Fido.search(time, instrument)
    assert isinstance(qr, UnifiedResponse)
    response = Fido.fetch(qr)
    assert len(response) == qr._numfile
    assert all(map(lambda x: x.endswith('.gz'), response))


def test_attr_reg():
    assert a.Instrument.gong == a.Instrument("GONG")
    assert a.ExtentType.synoptic == a.ExtentType("SYNOPTIC")


def test_client_repr(GSClient):
    """
    Repr check
    """
    output = str(GSClient)
    assert output[:50] == 'sunpy.net.dataretriever.sources.gong.GONGClient\n\nP'



"""
==============================================
Overlay an AIA image on a LASCO C2 coronagraph
==============================================

This example shows the steps needed to overlay the disc and off-limb
components of an AIA image within the masked occulter of a LASCO C2 image.
"""
from datetime import datetime

import hvpy
import matplotlib.pyplot as plt
from hvpy.datasource import DataSource

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
from sunpy.coordinates import SphericalScreen
from sunpy.map import Map
from sunpy.util.config import get_and_create_download_dir

###############################################################################
# First, we will acquire a calibrated LASCO C2 image from Helioviewer and
# create a map. hvpy uses the standard datetime instead of astropy.time.

lasco_jp2_file = hvpy.save_file(hvpy.getJP2Image(datetime(2011, 6, 7, 6, 34),
                                                 DataSource.LASCO_C2.value),
                                filename=get_and_create_download_dir() + "/LASCO_C2.jp2", overwrite=True)
lasco_map = Map(lasco_jp2_file)
aia_map = Map(sunpy.data.sample.AIA_171_IMAGE)

###############################################################################
# In order to plot off-limb features of the AIA image.
#
# Note that off-disk AIA data are not retained by default because an
# additional assumption is required to define the location of the AIA
# emission in 3D space. We can use `~sunpy.coordinates.SphericalScreen` to
# retain the off-disk AIA data. See
# :ref:`sphx_glr_generated_gallery_map_transformations_reprojection_spherical_screen.py`
# for more reference.

projected_coord = SkyCoord(0*u.arcsec, 0*u.arcsec,
                           obstime=lasco_map.observer_coordinate.obstime,
                           frame='helioprojective',
                           observer=lasco_map.observer_coordinate,
                           rsun=aia_map.coordinate_frame.rsun)
projected_header = sunpy.map.make_fitswcs_header(aia_map.data.shape,
                                                 projected_coord,
                                                 scale=u.Quantity(aia_map.scale),
                                                 instrument=aia_map.instrument,
                                                 wavelength=aia_map.wavelength)
# We use `~sunpy.coordinates.SphericalScreen` to ensure that the off limb AIA pixels
# are reprojected. Otherwise it will only be the on disk pixels that are reprojected.
with SphericalScreen(aia_map.observer_coordinate):
    aia_reprojected = aia_map.reproject_to(projected_header)

###############################################################################
# Finally, we plot the images by layering the AIA image on top of the LASCO C2
# image.

fig = plt.figure()
ax = fig.add_subplot(projection=lasco_map)
lasco_map.plot(axes=ax)
aia_reprojected.plot(axes=ax, clip_interval=(1, 99.9)*u.percent)
ax.set_title("AIA and LASCO C2 Overlay")

plt.show()



"""
ADAPT Map subclass definitions
"""


from sunpy.map.mapbase import GenericMap, SpatialPair
from sunpy.time import parse_time

__all__ = ['ADAPTMap']

class ADAPTMap(GenericMap):
    """
    ADAPT magnetic flux map

    The ADAPT (Air Force Data Assimilative Photospheric Flux Transport) flux transport model evolves an ensemble of realizations, using flux transport processes during periods for which there are no observations, and updates the ensemble using the ensemble least-squares (EnLS) data assimilation method to account for model and observational uncertainties.
    These data products are Carrington maps with an additional third dimension corresponding to the ensemble of models used to produce the ADAPT map.

    Here, the third dimension is ignored and only the first element of the third dimension is used to create the  resulting map.

    .. warning::

        Please note that there are several data arrays within the ADAPT FITS files and `~sunpy.map.Map` will, by default, try to read them all and fail.
        In these cases, you must specify the header-data Pair you want to read.
        For these data, it will always be the first one.
        You can specify this by passing the ``hdus`` keyword argument to `~sunpy.map.Map`

        .. code-block:: python

            >>> sunpy.map.Map("adapt40311_03k012_202401020800_i00005600n1.fts.gz", hdus=0)  # doctest: +SKIP

    References
    ----------
    * `NSO ADAPT Page <https://nso.edu/data/nisp-data/adapt-maps/>`__
    * `Data archive <https://gong.nso.edu/adapt/maps/>`__
    """
    @property
    def coordinate_system(self):
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1 == 'Long':
            ctype1 = 'CRLN-CAR'
        if ctype2 == 'Lat':
            ctype2 = 'CRLT-CAR'
        return SpatialPair(ctype1, ctype2)

    @property
    def date(self):
        return self._get_date('date-obs') or self._get_date('maptime') or super().date

    def _set_date(self, date):
        self.meta['date-obs'] = parse_time(date).utc.isot
        if 'maptime' in self.meta:
            del self.meta['maptime']

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an ADAPT map."""
        return header.get('model') == 'ADAPT'



# This test lives here because we have to exclude sunpy/data from pytest.
import os
from pathlib import Path

import sunpy.data.test


def mockreturn(path):
    paths = [
        (os.path.join('test', 'data', ''), (), ('code.py', 'test_file', 'code.pyc', '__init__.py'))
    ]
    return paths


def test_get_test_data_filenames(monkeypatch):
    monkeypatch.setattr(os, 'walk', mockreturn)
    monkeypatch.setattr(os.path, 'isfile', mockreturn)
    output = sunpy.data.test.get_test_data_filenames()
    assert isinstance(output, list)
    # Only the test file and not the py/pyc files should be in the return.
    assert output == [Path('test') / 'data' / 'test_file']



import pytest

from astropy.time import Time
from astropy.time.formats import erfa

# This registers the TimeTaiSeconds format with astropy
from sunpy.time import *  # NOQA


def test_time_t0():
    """
    Test that the offset at the epoch is zero. Note that the offset
    will be zero in both TAI and UTC
    """
    t = Time('1958-01-01 00:00:00', format='iso', scale='tai')
    assert t.tai.tai_seconds == 0.0
    with pytest.warns(erfa.ErfaWarning, match='dubious year'):
        assert t.utc.tai_seconds == 0.0


def test_tai_utc_offset():
    """
    Test that the offset between TAI and UTC is 10 s in 1972-01-01, when UTC
    was defined
    """
    t1 = Time('1972-01-01 00:00:00', scale='tai', format='iso')
    t2 = Time('1972-01-01 00:00:00', scale='utc', format='iso')
    assert t2.tai_seconds - t1.tai_seconds == 10.0


@pytest.mark.parametrize('time', [
    Time('1958-01-01T00:00:00', format='isot', scale='tai'),
    Time('1972-01-01T00:00:00', format='isot', scale='tai'),
    Time('2015-10-25T05:24:08', format='isot', scale='tai'),
    Time('2018-09-17T19:46:25', format='isot', scale='tai'),
])
def test_roundtrip(time):
    time_from_tai_seconds = Time(time.tai_seconds, scale='tai', format='tai_seconds')
    assert time_from_tai_seconds == time



from pathlib import Path

import astropy.units as u

import sunpy
from sunpy import config
from sunpy.data.data_manager.cache import Cache
from sunpy.data.data_manager.downloader import ParfiveDownloader
from sunpy.data.data_manager.manager import DataManager
from sunpy.data.data_manager.storage import SqliteStorage
from sunpy.util.config import CACHE_DIR

_download_dir = config.get('downloads', 'remote_data_manager_dir')


manager = DataManager(
    Cache(
        ParfiveDownloader(),
        SqliteStorage(_download_dir + '/data_manager.db'),
        _download_dir
    )
)
cache = Cache(
    ParfiveDownloader(),
    SqliteStorage(CACHE_DIR + '/cache.db'),
    CACHE_DIR,
    expiry=int(config.get('downloads', 'cache_expiry')) * u.day
)

__all__ = ["EARTH_IMAGE", "manager", "cache"]


# Low-resolution version of "The Blue Marble" (Earth as seen from Apollo 17)
EARTH_IMAGE = Path(__file__).parent / "The_Blue_Marble_150px.png"



"""
===================
Overlaying Two Maps
===================

This example demonstrates how to draw contours of one map on top of another
to compare features.
"""
import matplotlib.colors
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
from sunpy.map import Map

###############################################################################
# We start with the sample data. HMI shows the line-of-sight magnetic field at
# the photosphere while AIA 171 images show the resulting magnetic fields
# filled with hot plasma above, in the corona. We want to see what coronal
# features overlap with regions of strong line-of-sight magnetic fields.

aia_map = Map(sunpy.data.sample.AIA_171_IMAGE)
hmi_map = Map(sunpy.data.sample.HMI_LOS_IMAGE)

bottom_left = [0, 0] * u.arcsec
top_right = [800, 800] * u.arcsec
aia_smap = aia_map.submap(SkyCoord(*bottom_left,
                                   frame=aia_map.coordinate_frame),
                          top_right=SkyCoord(*top_right,
                                             frame=aia_map.coordinate_frame))
hmi_smap = hmi_map.submap(SkyCoord(*bottom_left,
                                   frame=hmi_map.coordinate_frame),
                          top_right=SkyCoord(*top_right,
                                             frame=hmi_map.coordinate_frame))

###############################################################################
# Let's set the contours of the HMI map from a few hundred to a thousand
# Gauss, which is the typical field strength associated with umbral regions of
# pores and sunspots.

levels = [-1000, -500, -250, 250, 500, 1000] * u.G

##############################################################################
# Now let's look at the result. Notice that we can see the coronal structures
# present on the AIA image and how they correspond to the line of sight
# magnetic field.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_smap)
aia_smap.plot(axes=ax)
hmi_smap.draw_contours(axes=ax, levels=levels, cmap="viridis")
norm = matplotlib.colors.Normalize(vmin=-1000, vmax=1000)
cbar = fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap="viridis"), ax=ax)
cbar.ax.set_title("Gauss", fontsize=10)
ax.set_title("AIA & HMI Composite Plot")

plt.show()



"""SDO Map subclass definitions"""

import numpy as np
from matplotlib.colors import CenteredNorm

import astropy.units as u
from astropy.coordinates import CartesianRepresentation, HeliocentricMeanEcliptic
from astropy.visualization import AsinhStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.map.mapbase import GenericMap, SpatialPair
from sunpy.map.sources.source_type import source_stretch
from sunpy.time import parse_time

__all__ = ['AIAMap', 'HMIMap', 'HMISynopticMap']


class AIAMap(GenericMap):
    """AIA Image Map.

    The Atmospheric Imaging Assembly is a set of four telescopes that employ
    normal-incidence, multi-layer coated optics to provide narrow-band imaging
    of the Sun. It provides high resolution full-disk images of the corona and
    transition region up to 0.5 solar radii above the solar limb with 1.5
    arcsecond angular resolution and 12-second temporal resolution. It observes
    the Sun in the following seven extreme ultraviolet bandpasses: 94 A
    (Fe XVIII), 131 A (Fe VIII, XXI), 171 A (Fe IX), 193 A (Fe XII, XXIV),
    211 A (Fe XIV), 304 A (He II), 335 A (Fe XVI). One telescope observes
    in the visible 1600 A (C IV) and the nearby continuum (1700 A).

    Notes
    -----
    Observer location: The standard AIA FITS header provides the spacecraft location in multiple
    coordinate systems, including Heliocentric Aries Ecliptic (HAE) and Heliographic Stonyhurst
    (HGS). SunPy uses the provided HAE coordinates due to accuracy concerns with the provided
    HGS coordinates, but other software packages may make different choices.

    References
    ----------
    * `SDO Mission Page <https://sdo.gsfc.nasa.gov/>`__
    * `Instrument Page <https://aia.lmsal.com>`__
    * `Fits Header keywords <http://jsoc.stanford.edu/doc/keywords/AIA/AIA02840_A_AIA-SDO_FITS_Keyword_Documents.pdf>`__
    * `Analysis Guide <https://www.lmsal.com/sdodocs/doc/dcur/SDOD0060.zip/zip/entry/>`__
    * Instrument Paper - :cite:t:`lemen_atmospheric_2012`
    * `wavelengths and temperature response reference <https://www.lmsal.com/sdodocs/doc/dcur/SDOD0060.zip/zip/entry/figures/aia_tel_resp.png>`__

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset

        import sunpy.map
        import astropy.units as u
        from sunpy.data.sample import AIA_171_IMAGE
        amap = sunpy.map.Map(AIA_171_IMAGE)
        amap.plot(clip_interval=(1.25, 99.5)*u.percent)
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        # Fill in some missing info
        self._nickname = self.detector
        self.plot_settings['cmap'] = self._get_cmap_name()
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, AsinhStretch(0.01)), clip=False)

    @property
    def _supported_observer_coordinates(self):
        return [(('haex_obs', 'haey_obs', 'haez_obs'), {'x': self.meta.get('haex_obs'),
                                                        'y': self.meta.get('haey_obs'),
                                                        'z': self.meta.get('haez_obs'),
                                                        'unit': u.m,
                                                        'representation_type': CartesianRepresentation,
                                                        'frame': HeliocentricMeanEcliptic})
                ] + super()._supported_observer_coordinates

    @property
    def observatory(self):
        """
        Returns the observatory.
        """
        return self.meta.get('telescop', '').split('/')[0]


    @property
    def reference_date(self):
        """
        The reference date for the coordinate system.

        DATE-OBS is derived from T_OBS by subtracting half the exposure time, so would not be a reference time.
        """
        return self._get_date('T_OBS') or super().reference_date

    def _set_reference_date(self, date):
        self.meta['t_obs'] = parse_time(date).utc.isot

    @property
    def detector(self):
        return self.meta.get("detector", "AIA")

    @property
    def unit(self):
        unit_str = self.meta.get('bunit', self.meta.get('pixlunit'))
        if unit_str is None:
            return

        return self._parse_fits_unit(unit_str)

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an AIA image"""
        return str(header.get('instrume', '')).startswith('AIA')


class HMIMap(GenericMap):
    """HMI Image Map.

    HMI consists of a refracting telescope, a polarization selector,
    an image stabilization system, a narrow band tunable filter
    and two 4096 pixel CCD cameras. It observes the full solar disk in the Fe I
    absorption line at 6173 Angstrom with a resolution of 1 arc-second.
    HMI takes images in a sequence of tuning and polarizations at a 4-second
    cadence for each camera. One camera is dedicated to a 45 s Doppler and
    line-of-sight field sequence while the other to a 90 s vector field
    sequence.

    References
    ----------
    * `SDO Mission Page <https://sdo.gsfc.nasa.gov/>`__
    * `Instrument Page <http://hmi.stanford.edu>`__
    * `Analysis Guide <http://hmi.stanford.edu/doc/magnetic/guide.pdf>`__

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset

        import sunpy.map
        from sunpy.data.sample import HMI_LOS_IMAGE
        hmi_map = sunpy.map.Map(HMI_LOS_IMAGE)
        hmi_map.plot()
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        if self.unit is not None and self.unit.is_equivalent(u.T):
            # Avoid JP2K images not having a norm due to UNIT8 data
            # This means they are not scaled correctly.
            if self.plot_settings.get('norm') is not None:
                # Magnetic field maps, not intensity maps
                self.plot_settings['norm'] = CenteredNorm()
        self._nickname = self.detector

    @property
    def waveunit(self):
        """
        The `~astropy.units.Unit` of the wavelength of this observation.

        Most HMI files seem to not have a parseable WAVEUNIT key so if it cannot be found
        we default to Angstrom
        """
        return super().waveunit or u.Angstrom

    @property
    def measurement(self):
        """
        Returns the measurement type.
        """
        content = self.meta.get('content', '').split(" ")
        if len(content) > 1 and content[0].lower() == 'hmi':
            # Sharp files have 'HMI' in the CONTENT header, but not all HMI files do.
            return content[1].lower()
        else:
            return content[0].lower()

    @property
    def observatory(self):
        """
        Returns the observatory.
        """
        return self.meta.get('telescop', '').split('/')[0]

    @property
    def reference_date(self):
        """
        The reference date for the coordinate system.

        DATE-OBS is derived from T_OBS by subtracting half the exposure time, so would not be a reference time.
        """
        return self._get_date('T_OBS') or super().reference_date

    def _set_reference_date(self, date):
        self.meta['T_OBS'] = parse_time(date).utc.isot

    @property
    def unit(self):
        unit_str = self.meta.get('bunit', None)
        if unit_str is None:
            return
        # Maxwells aren't in the IAU unit style manual and therefore not a valid FITS unit
        # The mapbase unit property forces this validation, so we must override it to prevent it.
        if (parsed_unit := u.Unit(unit_str)) == u.Unit('Mx/cm2'):
            return parsed_unit
        else:
            return super().unit

    @property
    def detector(self):
        return self.meta.get("detector", "HMI")

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an HMI image"""
        return (str(header.get('INSTRUME', '')).startswith('HMI') and
                not HMISynopticMap.is_datasource_for(data, header))


class HMISynopticMap(HMIMap):
    """
    SDO/HMI Synoptic Map.

    Synoptic maps are constructed from HMI 720s line-of-sight magnetograms
    collected over a 27-day solar rotation.

    See `~sunpy.map.sources.sdo.HMIMap` for information on the HMI instrument.

    Notes
    -----
    The sign of ``CDELT1`` in the header of (some) HMI synoptic maps is negative,
    but needs to be positive for the underlying data array in order to agree with
    HMI magnetograms as well as JSOC-hosted PNGs of the synoptic maps. Accordingly,
    we use the absolute value of ``CDELT1`` to force positivity.

    References
    ----------
    * `SDO Mission Page <https://sdo.gsfc.nasa.gov/>`__
    * `JSOC's HMI Synoptic Charts <http://jsoc.stanford.edu/HMI/LOS_Synoptic_charts.html>`__
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self.plot_settings['cmap'] = 'hmimag'
        self.plot_settings['norm'] = ImageNormalize(vmin=-1.5e3, vmax=1.5e3)

    @property
    def spatial_units(self):
        cunit1 = self.meta['cunit1']
        if cunit1 == 'Degree':
            cunit1 = 'deg'

        cunit2 = self.meta['cunit2']
        if cunit2 == 'Sine Latitude':
            cunit2 = 'deg'

        return SpatialPair(u.Unit(cunit1), u.Unit(cunit2))

    @property
    def scale(self):
        if self.meta['cunit2'] == 'Sine Latitude':
            # Since, this map uses the cylindrical equal-area (CEA) projection,
            # the spacing should be modified to 180/pi times the original value
            # Reference: Section 5.5, Thompson 2006
            return SpatialPair(np.abs(self.meta['cdelt1']) * self.spatial_units[0] / u.pixel,
                               180 / np.pi * self.meta['cdelt2'] * u.deg / u.pixel)

        return super().scale

    @property
    def date(self):
        """
        Image observation time.
        """
        return self._get_date('T_OBS') or super().date

    def _set_date(self, date):
        self.meta['T_OBS'] = parse_time(date).utc.isot

    @property
    def reference_date(self):
        """
        The reference date for the coordinate system.
        """
        return self._get_date('T_OBS') or super().reference_date

    def _set_reference_date(self, date):
        self.meta['T_OBS'] = parse_time(date).utc.isot

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """
        Determines if header corresponds to an HMI synoptic map.
        """
        return str(header.get('TELESCOP', '')).endswith('HMI') and 'carrington synoptic chart' in str(header.get('CONTENT', '')).lower()



import numpy as np

import astropy.units as u

from sunpy.net import attrs as a
from sunpy.net.dataretriever import GenericClient
from sunpy.net.dataretriever.client import QueryResponse
from sunpy.net.scraper import Scraper
from sunpy.time import TimeRange

__all__ = ["AIASynopsisClient"]


class AIASynopsisClient(GenericClient):
    """
    A client for retrieving AIA "synoptic" data from the JSOC.

    This dataset is not synoptic like HMI and MDI Synoptic images which are images of the solar surface reconstructed from many observations over a solar rotation but rather a synopsis of AIA data.

    The AIA synoptic data are calibrated Level 1.5 images with reduced 1k x 1k resolution at regular 2-minute cadence.

    Note
    ----

    This client does not support multiple wavelengths in a single query. If you want to download multiple wavelengths, you need to do it in separate queries.

    References
    ----------
    * `Readme <https://jsoc1.stanford.edu/data/aia/synoptic/README.html>`__

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a
    >>> results = Fido.search(a.Time("2016/1/1", "2016/1/1 00:01:00"),
    ...                       a.Instrument.aia, a.Level("1.5s"))  # doctest: +REMOTE_DATA
    >>> results  # doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    9 Results from the AIASynopsisClient:
    Source: https://jsoc1.stanford.edu/data/aia/synoptic/
    <BLANKLINE>
           Start Time               End Time        Instrument  Physobs  Source Provider Level Wavelength
    ----------------------- ----------------------- ---------- --------- ------ -------- ----- ----------
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S         94
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S        131
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S        171
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S        193
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S        211
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S        304
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S        335
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S       1600
    2016-01-01 00:00:00.000 2016-01-01 00:00:59.999        AIA intensity    SDO     JSOC  1.5S       4500
    <BLANKLINE>
    <BLANKLINE>
    """
    pattern = ('https://jsoc1.stanford.edu/data/aia/synoptic/'
               '{{year:4d}}/{{month:2d}}/{{day:2d}}/H{{hour:2d}}00/AIA{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}_{{Wavelength:04d}}.fits')
    required = {a.Time, a.Instrument, a.Level}

    @property
    def info_url(self):
        return "https://jsoc1.stanford.edu/data/aia/synoptic/"

    def search(self, *args, **kwargs):
        """
        Query this client for a list of results.

        Parameters
        ----------
        \\*args: `tuple`
            `sunpy.net.attrs` objects representing the query.
        \\*\\*kwargs: `dict`
            Any extra keywords to refine the search.

        Returns
        -------
        A `QueryResponse` instance containing the query result.
        """
        # TODO: There has to be a better way than repeating the entire search method.
        _, pattern, matchdict = self.pre_search_hook(*args, **kwargs)
        if "Wavelength" in matchdict:
            # The scarper uses string matching, so we have to convert the wavelength to a string
            # and remove the trailing zeros to match the pattern on the server.
            matchdict["Wavelength"] = str(np.round(matchdict["Wavelength"].min.to_value(u.AA, u.equivalencies.spectral()))).replace(".0","")
        scraper = Scraper(format=pattern)
        tr = TimeRange(matchdict['Start Time'], matchdict['End Time'])
        filesmeta = scraper._extract_files_meta(tr, matcher=matchdict)
        filesmeta = sorted(filesmeta, key=lambda k: k['url'])
        metalist = []
        if filesmeta:
            base_time = self.post_search_hook(filesmeta[0], matchdict)['Start Time']
            for i in filesmeta:
                rowdict = self.post_search_hook(i, matchdict)
                # Have to manually deal with sample rate if is it provided
                if sample_rate := matchdict.get("Sample"):
                    sample_rate = u.Quantity(sample_rate[0], u.s)
                    if (rowdict['Start Time'] - base_time).to(u.second) % sample_rate != 0 * u.s:
                        continue
                metalist.append(rowdict)
        return QueryResponse(metalist, client=self)

    @classmethod
    def register_values(cls):
        adict = {
            a.Instrument: [("AIA", "Data from the Atmospheric Imaging Assembly instrument.")],
            a.Physobs: [
                ("intensity", "Brightness or intensity of the solar atmosphere at different wavelengths.")
            ],
            a.Source: [("SDO", "The Solar Dynamics Observatory.")],
            a.Provider: [("JSOC", "Joint Science Operations Center at Stanford.")],
            a.Level: [("1.5s", "Level 1.5 data processed for quicker analysis.")],
        }
        return adict



from asdf_astropy.converters.coordinates.frame import FrameConverter


class SunpyFrameConverter(FrameConverter):
    def select_tag(self, obj, tags, ctx):
        # Sort the tags in reverse alphabetical order and pick the first (i.e.
        # the one with the highest version). This assumes that all the tags for
        # this converter are named the same other than the version number.
        tags = list(sorted(tags, reverse=True))
        return tags[0]


class HeliographicCarringtonConverter(SunpyFrameConverter):
    def from_yaml_tree(self, node, tag, ctx):
        # The 1.0.0 schema should be treated as having the observer at Earth
        if "1.0.0" in tag:
            node['frame_attributes']['observer'] = 'earth'
        return super().from_yaml_tree(node, tag, ctx)


SUNPY_FRAME_CONVERTERS = [
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/heliographic_stonyhurst-*",
        "sunpy.coordinates.frames.HeliographicStonyhurst"
    ),
    HeliographicCarringtonConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/heliographic_carrington-*",
        "sunpy.coordinates.frames.HeliographicCarrington"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/helioprojective-*",
        "sunpy.coordinates.frames.Helioprojective"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/helioprojectiveradial-*",
        "sunpy.coordinates.frames.HelioprojectiveRadial"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/heliocentricinertial-*",
        "sunpy.coordinates.frames.HeliocentricInertial"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/heliocentricearthecliptic-*",
        "sunpy.coordinates.frames.HeliocentricEarthEcliptic"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/heliocentric-*",
        "sunpy.coordinates.frames.Heliocentric"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/geocentricsolarecliptic-*",
        "sunpy.coordinates.frames.GeocentricSolarEcliptic"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/geocentricearthequatorial-*",
        "sunpy.coordinates.frames.GeocentricEarthEquatorial"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/geomagnetic-*",
        "sunpy.coordinates.frames.Geomagnetic"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/solarmagnetic-*",
        "sunpy.coordinates.frames.SolarMagnetic"
    ),
    SunpyFrameConverter(
        "tag:sunpy.org:sunpy/coordinates/frames/geocentricsolarmagnetospheric-*",
        "sunpy.coordinates.frames.GeocentricSolarMagnetospheric"
    ),
]



import os
import pathlib
from functools import singledispatchmethod
from collections import OrderedDict
from urllib.request import Request

import fsspec
import numpy as np

import astropy.io.fits
from astropy.utils.introspection import minversion
from astropy.wcs import WCS

from sunpy import log
from sunpy.data import cache
from sunpy.io._file_tools import detect_filetype, read_file
from sunpy.io._header import FileHeader
from sunpy.map.compositemap import CompositeMap
from sunpy.map.mapbase import GenericMap, MapMetaValidationError
from sunpy.map.mapsequence import MapSequence
from sunpy.util import expand_list
from sunpy.util.datatype_factory_base import (
    BasicRegistrationFactory,
    MultipleMatchError,
    NoMatchError,
    ValidationFunctionError,
)
from sunpy.util.exceptions import NoMapsInFileError, warn_user
from sunpy.util.io import expand_fsspec_open_file, is_uri, is_url, parse_path, possibly_a_path
from sunpy.util.metadata import MetaDict

SUPPORTED_ARRAY_TYPES = (np.ndarray,)
try:
    import dask.array

    SUPPORTED_ARRAY_TYPES += (dask.array.Array,)
except ImportError:
    pass

__all__ = ["Map", "MapFactory"]


class MapFactory(BasicRegistrationFactory):
    """
    A factory for generating coordinate aware 2D images.

    This factory takes a variety of inputs, such as file paths, wildcard
    patterns or (data, header) pairs.

    Depending on the input different return types are possible.

    Parameters
    ----------
    \\*inputs
        Inputs to parse for map objects. This can be one or more of the following:

        - A string or `~pathlib.Path` object pointing to a FITS file.
        - A directory containing FITS files (if there is more than one FITS file in the directory, it will return a list of Map objects).
        - A tuple containing a data array and a header (for modifying the data or header).
        - A `~sunpy.util.metadata.MetaDict` object, which includes data and metadata as a dictionary-like object.
        - An `astropy.wcs.WCS` object, which represents the World Coordinate System for the data.
        - A glob pattern to match multiple FITS files (e.g., ``eit_*.fits``).
        - A URL pointing to a FITS file (can be remote).
        - A combination of any of the above inputs.

    sequence : `bool`, optional
        Return a `sunpy.map.MapSequence` object comprised of all the parsed maps.

    composite : `bool`, optional
        Return a `sunpy.map.CompositeMap` object comprised of all the parsed maps.

    Returns
    -------
    `sunpy.map.GenericMap`
        If the input results in a singular map object, then that is returned.

    `list` of `~sunpy.map.GenericMap`
        If multiple inputs are given and ``sequence=False`` and ``composite=False``
        (the default) then a list of `~sunpy.map.GenericMap` objects will be
        returned.

    `sunpy.map.MapSequence`
        If the input corresponds to multiple maps and ``sequence=True`` is set,
        then a `~sunpy.map.MapSequence` object is returned.

    `sunpy.map.CompositeMap`
        If the input corresponds to multiple maps and ``composite=True`` is set,
        then a `~sunpy.map.CompositeMap` object is returned.

    Examples
    --------
    >>> import sunpy.map
    >>> from astropy.io import fits
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> mymap = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)  # doctest: +REMOTE_DATA +IGNORE_WARNINGS
    """

    def _read_file(self, fname, **kwargs):
        """
        Read in a file name and return the list of (data, meta) pairs in that file.
        """
        # File gets read here. This needs to be generic enough to seamlessly
        # call a fits file or a jpeg2k file, etc
        # NOTE: use os.fspath so that fname can be either a str or pathlib.Path
        # This can be removed once read_file supports pathlib.Path
        log.debug(f"Reading {fname}")
        try:
            filetype = detect_filetype(fname)
            if filetype == "asdf":
                import asdf
                if minversion(asdf, "3.1.0"):
                    _NO_MEMMAP_KWARGS = {"memmap": False, "lazy_load": False}
                else:
                    _NO_MEMMAP_KWARGS = {"copy_arrays": True, "lazy_load": False}
                with asdf.open(fname,** _NO_MEMMAP_KWARGS) as af:
                    pairs = [value for value in af.tree.values() if isinstance(value, GenericMap)]
                    return pairs
            else:
                pairs = read_file(os.fspath(fname), filetype=filetype, **kwargs)
        except Exception as e:
            msg = f"Failed to read {fname}\n{e}"
            if kwargs.get("allow_errors"):
                warn_user(msg)
                return []
            msg += "\n If you want to bypass these errors, pass `allow_errors=True`."
            raise OSError(msg) from e

        new_pairs = []
        for pair in pairs:
            filedata, filemeta = pair
            assert isinstance(filemeta, FileHeader)
            # This tests that the data is more than 1D
            if len(np.shape(filedata)) > 1:
                data = filedata
                meta = MetaDict(filemeta)
                new_pairs.append((data, meta))

        if not new_pairs:
            raise NoMapsInFileError(f"Found no HDUs with >= 2D data in '{fname}'.")

        return new_pairs

    def _validate_meta(self, meta):
        """
        Validate a meta argument.
        """
        if isinstance(meta, astropy.io.fits.header.Header):
            return True
        elif isinstance(meta, dict):
            return True
        else:
            return False

    def _parse_args(self, *args, allow_errors=False, **kwargs):
        """
        Parses an args list into data-header pairs.

        args can contain any mixture of the following entries:
        * tuples of data,header
        * data, header not in a tuple
        * data, wcs object in a tuple
        * data, wcs object not in a tuple
        * filename, as a str or pathlib.Path, which will be read
        * directory, as a str or pathlib.Path, from which all files will be read
        * glob, from which all files will be read
        * url, which will be downloaded and read
        * lists containing any of the above.

        Examples
        --------
        self._parse_args(data, header,
                         (data, header),
                         ['file1', 'file2', 'file3'],
                         'file4',
                         'directory1',
                         '*.fits')
        """
        # Account for nested lists of items
        args = expand_list(args)

        # Sanitise the input so that each 'type' of input corresponds to a different
        # class, so single dispatch can be used later
        parsed_args = []
        skip_next = False
        for i, arg in enumerate(args):
            if skip_next:
                skip_next = False
                continue

            if isinstance(arg, SUPPORTED_ARRAY_TYPES):
                # The next two items are data and a header
                data, header = args[i:i+2]
                parsed_args.append((data, header))
                skip_next = True
            elif isinstance(arg, str) and is_url(arg):
                # Replace URL string with a Request object to dispatch on later
                parsed_args.append(Request(arg))
            elif possibly_a_path(arg) and not is_uri(arg):
                # Replace path strings with Path objects
                parsed_args.append(pathlib.Path(arg))
            elif is_uri(arg):
                fsspec_kw = kwargs.get("fsspec_kwargs", {})
                # Get a list of open file objects (i.e. if given a glob)
                open_files = fsspec.open_files(arg, **fsspec_kw)
                # If any of the elements of the list are directories we want to
                # glob all files in them
                expanded_files = expand_list([expand_fsspec_open_file(f) for f in open_files])
                # Add all OpenFile objects to the parsed list
                parsed_args += expanded_files
            else:
                parsed_args.append(arg)

        # Parse the arguments
        # Note that this list can also contain GenericMaps if they are directly given to the factory
        data_header_pairs = []
        for arg in parsed_args:
            try:
                data_header_pairs += self._parse_arg(arg, allow_errors=allow_errors, **kwargs)
            except NoMapsInFileError as e:
                if not allow_errors:
                    raise
                warn_user(f"One of the arguments failed to parse with error: {e}")

        return data_header_pairs

    @singledispatchmethod
    def _parse_arg(self, arg, **kwargs):
        """
        Take a factory input and parse into (data, header) pairs.
        Must return a list, even if only one pair is returned.
        """
        raise ValueError(f"Invalid input: {arg!r}")

    @_parse_arg.register(tuple)
    def _parse_tuple(self, arg, **kwargs):
        # Data-header or data-WCS pair
        data, header = arg
        if isinstance(header, WCS):
            header = header.to_header()

        pair = data, header
        if self._validate_meta(header):
            pair = (data, OrderedDict(header))
        return [pair]

    @_parse_arg.register(GenericMap)
    def _parse_map(self, arg, **kwargs):
        return [arg]

    @_parse_arg.register(Request)
    def _parse_url(self, arg, **kwargs):
        url = arg.full_url
        path = str(cache.download(url).absolute())
        pairs = self._read_file(path, **kwargs)
        return pairs

    @_parse_arg.register(pathlib.Path)
    def _parse_path(self, arg, **kwargs):
        return parse_path(arg, self._read_file, **kwargs)

    @_parse_arg.register(fsspec.core.OpenFile)
    def _parse_fsspec_file(self, arg, **kwargs):
        # TODO: We should probably migrate the whole of the unified IO layer to
        # use fsspec for everything, but for now we parse the URI through
        return self._read_file(arg.full_name, **kwargs)

    def __call__(self, *args, composite=False, sequence=False, allow_errors=False, **kwargs):
        """Method for running the factory. Takes arbitrary arguments and
        keyword arguments and passes them to a sequence of pre-registered types
        to determine which is the correct Map-type to build.

        Arguments args and kwargs are passed through to the validation
        function and to the constructor for the final type. For Map types,
        validation function must take a data-header pair as an argument.

        Parameters
        ----------
        composite : `bool`, optional
            Indicates if collection of maps should be returned as a `~sunpy.map.CompositeMap`.
            Default is `False`.
        sequence : `bool`, optional
            Indicates if collection of maps should be returned as a `sunpy.map.MapSequence`.
            Default is `False`.
        allow_errors : `bool`, optional
            If set, bypass data-header pairs or files which cause an exception and warn instead.
            Defaults to `False`.

        Notes
        -----
        Extra keyword arguments are passed through to `sunpy.io._file_tools.read_file` such as
        ``memmap`` for FITS files.
        """
        data_header_pairs = self._parse_args(*args, allow_errors=allow_errors, **kwargs)
        new_maps = list()

        # Loop over each registered type and check to see if WidgetType
        # matches the arguments. If it does, use that type.
        for pair in data_header_pairs:
            if isinstance(pair, GenericMap):
                new_maps.append(pair)
                continue
            data, header = pair
            meta = MetaDict(header)

            try:
                new_map = self._check_registered_widgets(data, meta, **kwargs)
                new_maps.append(new_map)
            except (NoMatchError, MultipleMatchError, ValidationFunctionError, MapMetaValidationError) as e:
                if not allow_errors:
                    raise
                warn_user(f"One of the data, header pairs failed to validate with: {e}")

        if not len(new_maps):
            raise RuntimeError("No maps loaded")

        # If the list is meant to be a sequence, instantiate a map sequence
        if sequence:
            return MapSequence(new_maps, **kwargs)

        # If the list is meant to be a composite map, instantiate one
        if composite:
            return CompositeMap(new_maps, **kwargs)

        if len(new_maps) == 1:
            return new_maps[0]

        return new_maps

    def _check_registered_widgets(self, data, meta, **kwargs):
        candidate_widget_types = list()

        for key in self.registry:
            # Call the registered validation function for each registered class
            if self.registry[key](data, meta, **kwargs):
                candidate_widget_types.append(key)

        n_matches = len(candidate_widget_types)

        if n_matches == 0:
            if self.default_widget_type is None:
                raise NoMatchError("No types match specified arguments and no default is set.")
            else:
                candidate_widget_types = [self.default_widget_type]
        elif n_matches > 1:
            raise MultipleMatchError(
                "Too many candidate types identified "
                f"({candidate_widget_types}). "
                "Specify enough keywords to guarantee unique type "
                "identification."
            )

        # Only one is found
        WidgetType = candidate_widget_types[0]

        return WidgetType(data, meta, **kwargs)


class InvalidMapInput(ValueError):
    """Exception to raise when input variable is not a Map instance and does
    not point to a valid Map input file."""


class InvalidMapType(ValueError):
    """Exception to raise when an invalid type of map is requested with Map"""


class NoMapsFound(ValueError):
    """Exception to raise when input does not point to any valid maps or files"""


Map = MapFactory(registry=GenericMap._registry, default_widget_type=GenericMap,
                 additional_validation_functions=["is_datasource_for"])
